# Generated by Django 4.2.13 on 2024-07-10 12:34

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('openedx_ledger', '0011_deposit_models'),
    ]

    operations = [
        migrations.AlterField(
            model_name='deposit',
            name='sales_contract_reference_id',
            field=models.CharField(blank=True, db_index=True, help_text=('The reference ID for the specific sales contract which beget this deposit. The source system which ', 'stores the sales object is identified by the sales_contract_reference_provider field.'), max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='deposit',
            name='sales_contract_reference_provider',
            field=models.ForeignKey(blank=True, help_text='The system providing the source of truth for the sales contract which beget this deposit.', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='deposits', to='openedx_ledger.salescontractreferenceprovider'),
        ),
        migrations.AlterField(
            model_name='historicaldeposit',
            name='sales_contract_reference_id',
            field=models.CharField(blank=True, db_index=True, help_text=('The reference ID for the specific sales contract which beget this deposit. The source system which ', 'stores the sales object is identified by the sales_contract_reference_provider field.'), max_length=255, null=True),
        ),
    ]



# Generated by Django 3.2.19 on 2023-10-16 16:26

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import simple_history.models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('openedx_ledger', '0007_alter_externalfulfillmentprovider_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='HistoricalAdjustment',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', models.UUIDField(db_index=True, default=uuid.uuid4, editable=False)),
                ('adjustment_quantity', models.BigIntegerField(help_text='How many units to adjust for in the related transaction.')),
                ('reason', models.CharField(choices=[('unauthorized_enrollment', 'Unauthorized enrollment'), ('poor_content_fit', 'Poor content fit'), ('technical_challenges', 'Technical challenges'), ('missed_refund_or_date', 'Missed refund or date'), ('good_faith', 'Good faith/Relationship building')], db_index=True, default='technical_challenges', help_text='The primary reason for the existence of this adjustment.', max_length=255)),
                ('notes', models.TextField(blank=True, help_text='Any additional context you have for the existence of this adjustment.', null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('ledger', models.ForeignKey(blank=True, db_constraint=False, help_text='The Ledger instance with which this Adjustment is associated.', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='openedx_ledger.ledger')),
                ('transaction', models.ForeignKey(blank=True, db_constraint=False, help_text='The Transaction instance which adjusts the balance of the relevant ledger.', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='openedx_ledger.transaction')),
                ('transaction_of_interest', models.ForeignKey(blank=True, db_constraint=False, help_text='Any transaction of interest w.r.t. the reason for being of this adjustment. For example, the transaction of interest may point to some transaction record for which the enrolling user is unsatisfied, but for which we cannot issue a reversal due to business rules.', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='openedx_ledger.transaction')),
            ],
            options={
                'verbose_name': 'historical adjustment',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='Adjustment',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, unique=True)),
                ('adjustment_quantity', models.BigIntegerField(help_text='How many units to adjust for in the related transaction.')),
                ('reason', models.CharField(choices=[('unauthorized_enrollment', 'Unauthorized enrollment'), ('poor_content_fit', 'Poor content fit'), ('technical_challenges', 'Technical challenges'), ('missed_refund_or_date', 'Missed refund or date'), ('good_faith', 'Good faith/Relationship building')], db_index=True, default='technical_challenges', help_text='The primary reason for the existence of this adjustment.', max_length=255)),
                ('notes', models.TextField(blank=True, help_text='Any additional context you have for the existence of this adjustment.', null=True)),
                ('ledger', models.ForeignKey(help_text='The Ledger instance with which this Adjustment is associated.', on_delete=django.db.models.deletion.CASCADE, related_name='adjustments', to='openedx_ledger.ledger')),
                ('transaction', models.OneToOneField(help_text='The Transaction instance which adjusts the balance of the relevant ledger.', on_delete=django.db.models.deletion.CASCADE, related_name='adjustment', to='openedx_ledger.transaction')),
                ('transaction_of_interest', models.OneToOneField(blank=True, help_text='Any transaction of interest w.r.t. the reason for being of this adjustment. For example, the transaction of interest may point to some transaction record for which the enrolling user is unsatisfied, but for which we cannot issue a reversal due to business rules.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='adjustment_of_interest', to='openedx_ledger.transaction')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



"""
URLs for openedx_ledger.
"""
from django.contrib import admin
from django.urls import path, re_path  # pylint: disable=unused-import
from django.views.generic import TemplateView  # pylint: disable=unused-import

urlpatterns = [
    # TODO: Fill in URL patterns and views here.
    # re_path(r'', TemplateView.as_view(template_name="openedx_ledger/base.html")),
    path('admin/', admin.site.urls)
]



"""
Common utility functions that don't create/update/destroy business objects.
"""
import hashlib
from uuid import uuid4

from .constants import (
    INITIAL_DEPOSIT_TRANSACTION_SLUG,
    LEDGER_DEFAULT_IDEMPOTENCY_KEY_PREFIX,
    LEDGERED_SUBSIDY_IDEMPOTENCY_KEY_PREFIX,
)

TRANSACTION_METADATA_KEYS = {
    'lms_user_id',
    'content_key',
    'subsidy_access_policy_uuid',
}


def create_idempotency_key_for_ledger(subsidy_uuid=None):
    """
    Returns an idempotency key for creating a new ledger, given
    the UUID value of some Subsidy record.
    If subsidy_uuid is null, return some prefixed uuid4.
    """
    if subsidy_uuid:
        return f'{LEDGERED_SUBSIDY_IDEMPOTENCY_KEY_PREFIX}-{subsidy_uuid}'
    return f'{LEDGER_DEFAULT_IDEMPOTENCY_KEY_PREFIX}-{uuid4()}'


def create_idempotency_key_for_transaction(ledger, quantity, is_initial_deposit=False, **metadata):
    """
    Create a key that allows a transaction to be executed idempotently.
    """
    if is_initial_deposit:
        return f'{ledger.idempotency_key}-{quantity}-{INITIAL_DEPOSIT_TRANSACTION_SLUG}'

    idpk_data = {
        tx_key: value
        for tx_key, value in metadata.items()
        if tx_key in TRANSACTION_METADATA_KEYS
    }
    if not idpk_data:
        idpk_data = {
            'default_identifier': uuid4(),
        }
    hashed_metadata = hashlib.md5(str(idpk_data).encode()).hexdigest()
    return f'{ledger.idempotency_key}-{quantity}-{hashed_metadata}'



# Generated by Django 3.2.16 on 2023-02-16 16:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openedx_ledger', '0002_historicalledger_historicalreversal_historicaltransaction'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='historicaltransaction',
            name='content_uuid',
        ),
        migrations.RemoveField(
            model_name='transaction',
            name='content_uuid',
        ),
        migrations.AddField(
            model_name='historicalreversal',
            name='state',
            field=models.CharField(choices=[('created', 'Created'), ('pending', 'Pending'), ('committed', 'Committed')], db_index=True, default='created', max_length=255),
        ),
        migrations.AddField(
            model_name='historicaltransaction',
            name='content_key',
            field=models.CharField(blank=True, db_index=True, help_text='The globally unique content identifier.  Joinable with ContentMetadata.content_key in enterprise-catalog.', max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='historicaltransaction',
            name='reference_type',
            field=models.CharField(blank=True, choices=[('learner_credit_enterprise_course_enrollment_id', 'LearnerCreditEnterpriseCourseEnrollment ID')], db_index=True, help_text='The type of identifier used for `reference_id`.', max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='historicaltransaction',
            name='state',
            field=models.CharField(choices=[('created', 'Created'), ('pending', 'Pending'), ('committed', 'Committed')], db_index=True, default='created', max_length=255),
        ),
        migrations.AddField(
            model_name='historicaltransaction',
            name='subsidy_access_policy_uuid',
            field=models.UUIDField(blank=True, help_text='A reference to the subsidy access policy which was used to create a transaction for the content.', null=True),
        ),
        migrations.AddField(
            model_name='reversal',
            name='state',
            field=models.CharField(choices=[('created', 'Created'), ('pending', 'Pending'), ('committed', 'Committed')], db_index=True, default='created', max_length=255),
        ),
        migrations.AddField(
            model_name='transaction',
            name='content_key',
            field=models.CharField(blank=True, db_index=True, help_text='The globally unique content identifier.  Joinable with ContentMetadata.content_key in enterprise-catalog.', max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='transaction',
            name='reference_type',
            field=models.CharField(blank=True, choices=[('learner_credit_enterprise_course_enrollment_id', 'LearnerCreditEnterpriseCourseEnrollment ID')], db_index=True, help_text='The type of identifier used for `reference_id`.', max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='transaction',
            name='state',
            field=models.CharField(choices=[('created', 'Created'), ('pending', 'Pending'), ('committed', 'Committed')], db_index=True, default='created', max_length=255),
        ),
        migrations.AddField(
            model_name='transaction',
            name='subsidy_access_policy_uuid',
            field=models.UUIDField(blank=True, help_text='A reference to the subsidy access policy which was used to create a transaction for the content.', null=True),
        ),
        migrations.AlterField(
            model_name='historicaltransaction',
            name='reference_id',
            field=models.CharField(blank=True, db_index=True, help_text='The identifier of the item acquired via the transaction. e.g. a LearnerCreditEnterpriseCourseEnrollment ID.', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='reference_id',
            field=models.CharField(blank=True, db_index=True, help_text='The identifier of the item acquired via the transaction. e.g. a LearnerCreditEnterpriseCourseEnrollment ID.', max_length=255, null=True),
        ),
    ]



# Generated by Django 4.2.13 on 2024-06-28 00:26

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import simple_history.models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('openedx_ledger', '0010_transaction_parent_content_key'),
    ]

    operations = [
        migrations.CreateModel(
            name='SalesContractReferenceProvider',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(help_text='The pretty name of the contract reference provider.', max_length=255)),
                ('slug', models.SlugField(help_text='A unique slug for this sales contract reference provider.', max_length=32, unique=True)),
            ],
            options={
                'verbose_name': 'Sales Contract Reference Provider',
            },
        ),
        migrations.CreateModel(
            name='HistoricalDeposit',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', models.UUIDField(db_index=True, default=uuid.uuid4, editable=False)),
                ('desired_deposit_quantity', models.BigIntegerField(help_text='How many units to deposit via related transaction.')),
                ('sales_contract_reference_id', models.CharField(db_index=True, help_text=('The reference ID for the specific sales contract which beget this deposit. The source system which ', 'stores the sales object is identified by the sales_contract_reference_provider field.'), max_length=255)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('ledger', models.ForeignKey(blank=True, db_constraint=False, help_text='The Ledger instance with which this deposit is associated.', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='openedx_ledger.ledger')),
                ('sales_contract_reference_provider', models.ForeignKey(blank=True, db_constraint=False, help_text='The system providing the source of truth for the sales contract which beget this deposit.', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='openedx_ledger.salescontractreferenceprovider')),
                ('transaction', models.ForeignKey(blank=True, db_constraint=False, help_text='The Transaction instance which represents this deposit and influences the balance of the relevant ledger.', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='openedx_ledger.transaction')),
            ],
            options={
                'verbose_name': 'historical deposit',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='Deposit',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, unique=True)),
                ('desired_deposit_quantity', models.BigIntegerField(help_text='How many units to deposit via related transaction.')),
                ('sales_contract_reference_id', models.CharField(db_index=True, help_text=('The reference ID for the specific sales contract which beget this deposit. The source system which ', 'stores the sales object is identified by the sales_contract_reference_provider field.'), max_length=255)),
                ('ledger', models.ForeignKey(help_text='The Ledger instance with which this deposit is associated.', on_delete=django.db.models.deletion.CASCADE, related_name='deposits', to='openedx_ledger.ledger')),
                ('sales_contract_reference_provider', models.ForeignKey(help_text='The system providing the source of truth for the sales contract which beget this deposit.', on_delete=django.db.models.deletion.CASCADE, related_name='deposits', to='openedx_ledger.salescontractreferenceprovider')),
                ('transaction', models.OneToOneField(help_text='The Transaction instance which represents this deposit and influences the balance of the relevant ledger.', on_delete=django.db.models.deletion.CASCADE, related_name='deposit', to='openedx_ledger.transaction')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



"""
Test factories for openedx-ledger models.
"""
from uuid import uuid4

import factory

from openedx_ledger.models import (
    Adjustment,
    Deposit,
    ExternalFulfillmentProvider,
    ExternalTransactionReference,
    Ledger,
    Reversal,
    SalesContractReferenceProvider,
    Transaction,
    TransactionStateChoices,
    UnitChoices,
)


class LedgerFactory(factory.django.DjangoModelFactory):
    """
    Test factory for the `Ledger` model.

    By default, no transactions are created for this test Ledger.
    """
    class Meta:
        model = Ledger

    uuid = factory.LazyFunction(uuid4)
    idempotency_key = factory.Faker("lexify", text="subsidy-????")
    unit = UnitChoices.USD_CENTS


class TransactionFactory(factory.django.DjangoModelFactory):
    """
    Test factory for the `Transaction` model.
    """
    class Meta:
        model = Transaction

    uuid = factory.LazyFunction(uuid4)
    idempotency_key = factory.LazyFunction(uuid4)
    state = TransactionStateChoices.COMMITTED
    quantity = factory.Faker("random_int", min=-100000, max=-100)
    ledger = factory.Iterator(Ledger.objects.all())
    lms_user_id = factory.Faker("random_int", min=1, max=1000)
    lms_user_email = factory.Faker("email")
    content_key = factory.LazyAttribute(lambda tx: f"course-v1:{tx.parent_content_key}+2023")
    parent_content_key = factory.Faker("lexify", text="???+?????101")
    content_title = factory.Faker("lexify", text="???: ?????? ???")


class ExternalFulfillmentProviderFactory(factory.django.DjangoModelFactory):
    """
    Test factory for the `ExternalFulfillmentProvider` model.
    """
    class Meta:
        model = ExternalFulfillmentProvider

    name = factory.Faker("company")
    slug = factory.Faker("slug")


class ExternalTransactionReferenceFactory(factory.django.DjangoModelFactory):
    """
    Test factory for the `ExternalTransactionReferenceFactory` model.
    """
    class Meta:
        model = ExternalTransactionReference

    external_reference_id = factory.Faker("lexify", text="????-????-????-????")
    transaction = factory.Iterator(Transaction.objects.all())
    external_fulfillment_provider = factory.Iterator(ExternalFulfillmentProvider.objects.all())


class ReversalFactory(factory.django.DjangoModelFactory):
    """
    Test factory for the `Reversal` model.
    """
    class Meta:
        model = Reversal

    uuid = factory.LazyFunction(uuid4)
    transaction = factory.Iterator(Transaction.objects.all())
    idempotency_key = factory.LazyFunction(uuid4)
    state = TransactionStateChoices.COMMITTED
    quantity = factory.Faker("random_int", min=100, max=10000)


class AdjustmentFactory(factory.django.DjangoModelFactory):
    """
    Test factory for the `Adjustment` model.
    """
    class Meta:
        model = Adjustment

    ledger = factory.SubFactory(LedgerFactory)
    transaction = factory.SubFactory(
        TransactionFactory,
        ledger=factory.SelfAttribute('..ledger'),
        quantity=factory.SelfAttribute('..adjustment_quantity'),
    )
    adjustment_quantity = factory.Faker("random_int", min=100, max=10000)


class SalesContractReferenceProviderFactory(factory.django.DjangoModelFactory):
    """
    Test factory for the `SalesContractReferenceProvider` model.
    """
    class Meta:
        model = SalesContractReferenceProvider

    name = factory.Faker("company")
    slug = factory.Faker("slug")


class DepositFactory(factory.django.DjangoModelFactory):
    """
    Test factory for the `Deposit` model.
    """
    class Meta:
        model = Deposit

    ledger = factory.SubFactory(LedgerFactory)
    transaction = factory.SubFactory(
        TransactionFactory,
        ledger=factory.SelfAttribute('..ledger'),
        quantity=factory.SelfAttribute('..desired_deposit_quantity'),
    )
    desired_deposit_quantity = factory.Faker("random_int", min=100, max=10000)
    sales_contract_reference_id = factory.Faker("lexify", text="????-????-????-????")
    sales_contract_reference_provider = factory.SubFactory(SalesContractReferenceProviderFactory)



"""
Tests for the utils.py module.
"""
from itertools import chain, combinations

from django.test import TestCase

from openedx_ledger import utils
from openedx_ledger.test_utils.factories import LedgerFactory


def powerset(iterable):
    """
    https://docs.python.org/3/library/itertools.html#itertools-recipes
    powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)

    Be warned, this gets very big, very fast, with higher
    cardinatlities of `iterable`.  This function asserts that
    no more than 7 things are provided as input.
    """
    s = list(iterable)
    assert len(s) < 8, 'This powerset will be too big'
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))


class TransactionIdempotencyKeyTests(TestCase):
    """
    Tests for the create_idempotency_key_for_transaction function.
    """
    def test_transaction_idempotency_key_varies(self):
        """
        Test that we get unique idempotency keys for
        transactions when changing any of the possible
        subsets of keys against some baseline transaction.
        """
        ledger = LedgerFactory()
        baseline_metadata = {
            key: f'{key}-value-1'
            for key in utils.TRANSACTION_METADATA_KEYS
        }
        baseline_key = utils.create_idempotency_key_for_transaction(
            ledger, 10, **baseline_metadata,
        )

        combos = list(powerset(utils.TRANSACTION_METADATA_KEYS))
        generated_keys = set()
        for combo in combos:
            comparison_metadata = baseline_metadata.copy()
            comparison_metadata.update({
                key: f'{key}-value-2'
                for key in combo
            })
            generated_key = utils.create_idempotency_key_for_transaction(
                ledger=ledger, quantity=10, **comparison_metadata,
            )
            if combo:
                assert baseline_key != generated_key
            else:
                assert baseline_key == generated_key
            generated_keys.add(generated_key)

        assert len(combos) == len(generated_keys)

    def test_initial_deposit_transaction(self):
        """
        Test we get an expected key for the initial transaction of a ledger.
        """
        ledger = LedgerFactory(idempotency_key='ledger-key')

        key = utils.create_idempotency_key_for_transaction(
            ledger, quantity=100, is_initial_deposit=True,
        )
        assert 'ledger-key-100-initial-deposit' == key



"""
Admin configuration for openedx_ledger models.
"""
from django import forms
from django.conf import settings
from django.contrib import admin
from django.http import HttpResponseRedirect
from django.urls import re_path, reverse
from django_object_actions import DjangoObjectActions
from simple_history.admin import SimpleHistoryAdmin

from openedx_ledger import api, constants, models, views


def can_modify():
    return getattr(settings, 'ALLOW_LEDGER_MODIFICATION', False)


def cents_to_usd_string(balance_in_cents):
    """
    Helper to convert cents as in int to dollars as a
    nicely formatted string.
    """
    return "${:,.2f}".format(float(balance_in_cents) / constants.CENTS_PER_US_DOLLAR)


@admin.register(models.Ledger)
class LedgerAdmin(SimpleHistoryAdmin):
    """
    Admin configuration for the Ledger model.
    """

    class Meta:
        """
        Metaclass for LedgerAdmin.
        """
        model = models.Ledger

    fields = ('uuid', 'idempotency_key', 'unit', 'balance_usd', 'metadata')
    # The autocomplete_fields of AdjustmentAdmin include the ledger field,
    # which in turn requires that we define here that ledgers are
    # searchable by uuid.
    search_fields = [
        'uuid',
    ]
    if can_modify():
        readonly_fields = ('uuid', 'balance_usd')
    else:
        readonly_fields = fields

    # Do not add balance here, it's a computed value.
    list_display = ('uuid', 'unit', 'idempotency_key')

    def balance_usd(self, obj):
        """
        Passthrough function to calculate the ledger balance.
        """
        return cents_to_usd_string(obj.balance())


@admin.register(models.ExternalFulfillmentProvider)
class ExternalFulfillmentProviderAdmin(SimpleHistoryAdmin):
    """
    Admin configuration for the ExternalFulfillmentProvider model.
    """
    class Meta:
        """
        Metaclass for ExternalFulfillmentProviderAdmin.
        """

        model = models.ExternalFulfillmentProvider
        fields = '__all__'

    search_fields = ('name', 'slug',)
    list_display = ('name', 'slug',)


class ExternalTransactionReferenceInlineAdmin(admin.TabularInline):
    """
    Inline admin configuration for the ExternalTransactionReference model.
    """
    model = models.ExternalTransactionReference


class AdjustmentInlineAdmin(admin.TabularInline):
    """
    Inline admin configuration for the Adjustment model.
    """
    model = models.Adjustment
    fk_name = 'transaction'
    fields = [
        'uuid',
        'get_quantity_usd',
        'reason',
        'created',
        'modified',
    ]
    readonly_fields = fields
    show_change_link = True

    @admin.display(description='Amount in U.S. Dollars')
    def get_quantity_usd(self, obj):
        if not obj._state.adding:  # pylint: disable=protected-access
            return cents_to_usd_string(obj.adjustment_quantity)
        return None


class TransactionTypeFilter(admin.SimpleListFilter):
    """
    Filter transaction by "type".

    I'm loosely defining "type" as one of: deposit, adjustment or spend. In turn, "spend" is defined as any transaction
    that isn't a deposit or an adjustment.
    """
    title = 'transaction type'
    parameter_name = 'transaction type'

    def lookups(self, request, model_admin):
        return [
            ('spend', 'Spend'),
            ('adjustment', 'Adjustment'),
            ('deposit', 'Deposit'),
        ]

    def queryset(self, request, queryset):
        if self.value() == 'spend':
            return queryset.filter(deposit__isnull=True, adjustment__isnull=True)
        elif self.value() == 'adjustment':
            return queryset.filter(adjustment__isnull=False)
        elif self.value() == 'deposit':
            return queryset.filter(deposit__isnull=False)
        else:
            return queryset


@admin.register(models.Transaction)
class TransactionAdmin(DjangoObjectActions, SimpleHistoryAdmin):
    """
    Admin configuration for the Transaction model.
    """

    class Meta:
        """
        Metaclass for TransactionAdmin.
        """

        model = models.Transaction
        fields = '__all__'

    search_fields = (
        'content_key',
        'lms_user_id',
        'uuid',
        'external_reference__external_reference_id',
        'subsidy_access_policy_uuid',
    )
    _all_fields = [
        field.name for field in models.Transaction._meta.get_fields()
        if field.name not in {'external_reference', 'adjustment', 'adjustment_of_interest'}
    ]
    _writable_fields = [
        'fulfillment_identifier',
    ]
    list_display = (
        'uuid',
        'lms_user_id',
        'content_key',
        'subsidy_access_policy_uuid',
        'quantity',
        'state',
        'modified',
        'has_reversal',
    )
    list_filter = (
        'state',
        TransactionTypeFilter,
    )

    if can_modify():
        readonly_fields = [
            field_name for field_name in _all_fields
            if field_name not in ['fulfillment_identifier']
        ]
    else:
        readonly_fields = _all_fields

    def get_inlines(self, request, obj):
        if obj and not obj._state.adding:  # pylint: disable=protected-access
            return [AdjustmentInlineAdmin, ExternalTransactionReferenceInlineAdmin]
        return [ExternalTransactionReferenceInlineAdmin]

    @admin.display(ordering='reversal', description='Has a reversal')
    def has_reversal(self, obj):
        return bool(obj.get_reversal())

    change_actions = ('reverse_transaction',)

    def get_queryset(self, request):
        queryset = super().get_queryset(request)
        return queryset.select_related('reversal')

    @admin.action(
        description="Reverse a transaction and unenroll the learner from the platform representation of the course."
    )
    def reverse_transaction(self, request, obj):
        """
        Redirect to the reverse transaction view.
        """
        # url names coming from get_urls are prefixed with 'admin' namespace
        reverse_transaction_url = reverse("admin:reverse_transaction", args=(obj.uuid,))
        return HttpResponseRedirect(reverse_transaction_url)

    reverse_transaction.label = "Unenroll & Refund"

    def get_urls(self):
        """
        Returns the additional urls used by DjangoObjectActions.
        """
        customer_urls = [
            re_path(
                r"^([^/]+)/reverse_transaction$",
                self.admin_site.admin_view(views.ReverseTransactionView.as_view()),
                name="reverse_transaction"
            )
        ]
        return customer_urls + super().get_urls()


@admin.register(models.Reversal)
class ReversalAdmin(SimpleHistoryAdmin):
    """
    Admin configuration for the Reversal model.
    """

    class Meta:
        """
        Metaclass for ReversalAdmin.
        """

        model = models.Reversal
        fields = '__all__'

    autocomplete_fields = [
        'transaction',
    ]
    list_display = (
        'uuid',
        'quantity',
        'state',
        'transaction',
        'modified',
    )
    ordering = ('-modified',)


class AdjustmentAdminCreateForm(forms.ModelForm):
    """
    Form that allows users to enter adjustment quantities in dollars
    instead of cents.
    """
    class Meta:
        model = models.Adjustment
        fields = [
            'ledger',
            'quantity_usd_input',
            'reason',
            'notes',
            'transaction_of_interest',
        ]

    quantity_usd_input = forms.FloatField(
        required=True,
        help_text='Amount of adjustment in US Dollars. If the adjustment reduces the total deposits, please verify \
        that the policies\' spend limit associated to this subsidy does not exceed the new total deposits after this \
        adjustment is made.',
    )


class AdjustmentAdminChangeForm(forms.ModelForm):
    """
    Form for reading and changing only the allowed fields of an existing adjustment record.
    """
    class Meta:
        model = models.Adjustment
        fields = [
            'ledger',
            'reason',
            'notes',
            'transaction_of_interest',
            'transaction',
            'adjustment_quantity',
        ]


@admin.register(models.Adjustment)
class AdjustmentAdmin(SimpleHistoryAdmin):
    """
    Admin configuration for the Adjustment model.
    """
    form = AdjustmentAdminCreateForm

    _readonly_fields = [
        'get_quantity_usd',
        'uuid',
        'transaction',
        'adjustment_quantity',
        'created',
        'modified',
    ]

    list_display = (
        'uuid',
        'get_ledger_uuid',
        'get_quantity_usd',
        'reason',
        'created',
        'modified',
    )
    list_filter = (
        'reason',
    )
    autocomplete_fields = [
        'ledger',
        'transaction_of_interest',
    ]

    def get_readonly_fields(self, request, obj=None):
        """
        Don't allow changing the ledger if we've already saved the adjustment record.
        """
        if obj and not obj._state.adding:  # pylint: disable=protected-access
            return ['ledger'] + self._readonly_fields
        return self._readonly_fields

    def get_fields(self, request, obj=None):
        """
        Don't include the ``quantity_usd_input`` field unless we're creating a new adjustment.
        """
        if not obj:
            # When we're adding a new adjustment, show only the fields relevant for input.
            return [
                'ledger',
                'quantity_usd_input',
                'reason',
                'notes',
                'transaction_of_interest',
            ]
        else:
            # When we're reading/editing an existing adjustment, show all fields EXCEPT for the fields exclusively used
            # for entering input (quantity_usd_input).
            return [
                field for field in super().get_fields(request, obj)
                if field != 'quantity_usd_input'
            ]

    def get_form(self, request, obj=None, **kwargs):  # pylint: disable=arguments-differ
        """
        Don't worry about validating the ``quantity_usd_input`` unless we're creating a new adjustment.
        """
        if obj and not obj._state.adding:  # pylint: disable=protected-access
            kwargs['form'] = AdjustmentAdminChangeForm
        return super().get_form(request, obj, **kwargs)

    @admin.display(description='Amount in U.S. Dollars')
    def get_quantity_usd(self, obj):
        if not obj._state.adding:  # pylint: disable=protected-access
            return cents_to_usd_string(obj.adjustment_quantity)
        return None

    @admin.display(ordering='uuid', description='Ledger uuid')
    def get_ledger_uuid(self, obj):
        return obj.ledger.uuid

    def save_model(self, request, obj, form, change):
        if change:
            super().save_model(request, obj, form, change)
        else:
            raw_usd_input = form.cleaned_data.get('quantity_usd_input')
            quantity_usd_cents = raw_usd_input * constants.CENTS_PER_US_DOLLAR
            # AED 2023-10-16: Use the auto-generated "stub" UUID for the Adjustment record
            # to persist the Adjustment record, so that Django Admin doesn't get lost
            # when a user clicks "Save and Continue Editing".
            api.create_adjustment(
                adjustment_uuid=obj.uuid,
                ledger=obj.ledger,
                quantity=quantity_usd_cents,
                reason=obj.reason,
                notes=obj.notes,
                transaction_of_interest=obj.transaction_of_interest,
            )


@admin.register(models.SalesContractReferenceProvider)
class SalesContractReferenceProviderAdmin(SimpleHistoryAdmin):
    """
    Admin configuration for the SalesContractReferenceProvider model.
    """
    class Meta:
        """
        Metaclass for SalesContractReferenceProviderAdmin.
        """

        model = models.SalesContractReferenceProvider
        fields = '__all__'

    search_fields = ('name', 'slug',)
    list_display = ('name', 'slug',)


class DepositAdminCreateForm(forms.ModelForm):
    """
    Form that allows users to enter deposit quantities in dollars
    instead of cents.
    """
    class Meta:
        model = models.Deposit
        fields = [
            'ledger',
            'quantity_usd_input',
            'sales_contract_reference_id',
            'sales_contract_reference_provider',
        ]

    quantity_usd_input = forms.FloatField(
        required=True,
        help_text=(
            "Amount of deposit in US Dollars. Remember to increase the related policy spend limits to take advantage "
            "of this new added value."
        ),
    )


class DepositAdminChangeForm(forms.ModelForm):
    """
    Form for reading and changing only the allowed fields of an existing Deposit record.
    """
    class Meta:
        model = models.Deposit
        fields = [
            'ledger',
            'transaction',
            'desired_deposit_quantity',
            'sales_contract_reference_id',
            'sales_contract_reference_provider',
        ]


@admin.register(models.Deposit)
class DepositAdmin(SimpleHistoryAdmin):
    """
    Admin configuration for the Deposit model.
    """
    form = DepositAdminCreateForm

    _readonly_fields = [
        'get_quantity_usd',
        'uuid',
        'transaction',
        'desired_deposit_quantity',
        'created',
        'modified',
    ]

    list_display = (
        'uuid',
        'get_ledger_uuid',
        'get_quantity_usd',
        'sales_contract_reference_id',
        'created',
        'modified',
    )

    # There is no low-cardinality field to filter on for Deposits.
    # list_filter = ()

    autocomplete_fields = [
        'ledger',
    ]

    def get_readonly_fields(self, request, obj=None):
        """
        Don't allow changing the ledger if we've already saved the deposit record.
        """
        if obj and not obj._state.adding:  # pylint: disable=protected-access
            return ['ledger'] + self._readonly_fields
        return self._readonly_fields

    def get_fields(self, request, obj=None):
        """
        Don't include the ``quantity_usd_input`` field unless we're creating a new deposit.
        """
        # When we're adding a new deposit, use default fields
        if not obj:
            return [
                'ledger',
                'quantity_usd_input',
                'sales_contract_reference_id',
                'sales_contract_reference_provider',
            ]
        else:
            # Don't show the USD amount input field on read/change
            return [
                field for field in super().get_fields(request, obj)
                if field != 'quantity_usd_input'
            ]

    def get_form(self, request, obj=None, **kwargs):  # pylint: disable=arguments-differ
        """
        Don't worry about validating the ``quantity_usd_input`` unless we're creating a new deposit.
        """
        if obj and not obj._state.adding:  # pylint: disable=protected-access
            kwargs['form'] = DepositAdminChangeForm
        return super().get_form(request, obj, **kwargs)

    @admin.display(description='Amount in U.S. Dollars')
    def get_quantity_usd(self, obj):
        if not obj._state.adding:  # pylint: disable=protected-access
            return cents_to_usd_string(obj.desired_deposit_quantity)
        return None

    @admin.display(ordering='uuid', description='Ledger uuid')
    def get_ledger_uuid(self, obj):
        return obj.ledger.uuid

    def save_model(self, request, obj, form, change):
        if change:
            super().save_model(request, obj, form, change)
        else:
            raw_usd_input = form.cleaned_data.get('quantity_usd_input')
            quantity_usd_cents = raw_usd_input * constants.CENTS_PER_US_DOLLAR
            # AED 2023-10-16: Use the auto-generated "stub" UUID for the Adjustment record
            # to persist the Adjustment record, so that Django Admin doesn't get lost
            # when a user clicks "Save and Continue Editing".
            api.create_deposit(
                deposit_uuid=obj.uuid,
                ledger=obj.ledger,
                quantity=quantity_usd_cents,
                sales_contract_reference_id=obj.sales_contract_reference_id,
                sales_contract_reference_provider=obj.sales_contract_reference_provider,
            )



# Generated by Django 3.2.23 on 2024-02-15 16:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openedx_ledger', '0009_add_email_and_title_to_transactions'),
    ]

    operations = [
        migrations.AddField(
            model_name='historicaltransaction',
            name='parent_content_key',
            field=models.CharField(blank=True, db_index=True, help_text='Identifier for the parent of the content_key.  Also should be joinable with ContentMetadata.content_key in enterprise-catalog. In practice, this is the course key corresponding to a course run key.', max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='transaction',
            name='parent_content_key',
            field=models.CharField(blank=True, db_index=True, help_text='Identifier for the parent of the content_key.  Also should be joinable with ContentMetadata.content_key in enterprise-catalog. In practice, this is the course key corresponding to a course run key.', max_length=255, null=True),
        ),
    ]



# Generated by Django 3.2.18 on 2023-03-23 12:51

from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('openedx_ledger', '0004_ledger_idpkey_blankable'),
    ]

    operations = [
        migrations.AlterField(
            model_name='historicalledger',
            name='idempotency_key',
            field=models.CharField(blank=True, db_index=True, help_text='An idempotency key is a unique value generated by the client of the Ledger API which the Ledger and Transaction APIs use to recognize subsequent retries of the same redemption (i.e. a Transaction that leads to the fulfillment of an enrollment or entitlement for a particular user id in a particular content key.We suggest incorporating V4 UUIDs along with sufficiently unique or random values representing the desired redemption to avoid unintended collisions.Utility methods are provided in the ``utils.py`` module to help clients generate appropriate idempotency keys.', max_length=255),
        ),
        migrations.AlterField(
            model_name='historicalledger',
            name='metadata',
            field=jsonfield.fields.JSONField(blank=True, help_text='Any additionaly metadata that a client may want to associate with this Ledger instance.', null=True),
        ),
        migrations.AlterField(
            model_name='historicalledger',
            name='unit',
            field=models.CharField(choices=[('usd_cents', 'U.S. Dollar (Cents)'), ('seats', 'Seats in a course')], db_index=True, default='usd_cents', help_text='The unit in which Transaction quantities related to this Ledger instance are denominated.', max_length=255),
        ),
        migrations.AlterField(
            model_name='historicalreversal',
            name='idempotency_key',
            field=models.CharField(db_index=True, help_text="An idempotency key is a unique value generated by the client of the Ledger API which the Ledger and Transaction APIs use to recognize subsequent retries of the same redemption (i.e. a Transaction that leads to the fulfillment of an enrollment or entitlement for a particular user id in a particular content key).We suggest incorporating V4 UUIDs along with sufficiently unique or random values representing the desired redemption to avoid unintended collisions.In particular, a Transaction's idempotency_key should incorporate it's corresponding ledger's idempotency_key.Utility methods are provided in the ``utils.py`` module to help clients generate appropriate idempotency keys.", max_length=255),
        ),
        migrations.AlterField(
            model_name='historicalreversal',
            name='metadata',
            field=jsonfield.fields.JSONField(blank=True, help_text='Any additionaly metadata that a client may want to associate with this Transaction.', null=True),
        ),
        migrations.AlterField(
            model_name='historicalreversal',
            name='quantity',
            field=models.BigIntegerField(help_text='How many units (as defined in the associated Ledger instance) this Transaction represents.'),
        ),
        migrations.AlterField(
            model_name='historicalreversal',
            name='state',
            field=models.CharField(choices=[('created', 'Created'), ('pending', 'Pending'), ('committed', 'Committed'), ('failed', 'Failed')], db_index=True, default='created', help_text="The current state of the Transaction. One of: ['Created', 'Pending', 'Committed', 'Failed']", max_length=255),
        ),
        migrations.AlterField(
            model_name='historicalreversal',
            name='transaction',
            field=models.ForeignKey(blank=True, db_constraint=False, help_text='The Transaction instance which is reversed by this Reversal instance.', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='openedx_ledger.transaction'),
        ),
        migrations.AlterField(
            model_name='historicaltransaction',
            name='idempotency_key',
            field=models.CharField(db_index=True, help_text="An idempotency key is a unique value generated by the client of the Ledger API which the Ledger and Transaction APIs use to recognize subsequent retries of the same redemption (i.e. a Transaction that leads to the fulfillment of an enrollment or entitlement for a particular user id in a particular content key).We suggest incorporating V4 UUIDs along with sufficiently unique or random values representing the desired redemption to avoid unintended collisions.In particular, a Transaction's idempotency_key should incorporate it's corresponding ledger's idempotency_key.Utility methods are provided in the ``utils.py`` module to help clients generate appropriate idempotency keys.", max_length=255),
        ),
        migrations.AlterField(
            model_name='historicaltransaction',
            name='ledger',
            field=models.ForeignKey(blank=True, db_constraint=False, help_text='The Ledger instance with which this Transaction is associated.', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='openedx_ledger.ledger'),
        ),
        migrations.AlterField(
            model_name='historicaltransaction',
            name='lms_user_id',
            field=models.IntegerField(blank=True, db_index=True, help_text='The id of the Open edX LMS user record with which this Transaction is associated.', null=True),
        ),
        migrations.AlterField(
            model_name='historicaltransaction',
            name='metadata',
            field=jsonfield.fields.JSONField(blank=True, help_text='Any additionaly metadata that a client may want to associate with this Transaction.', null=True),
        ),
        migrations.AlterField(
            model_name='historicaltransaction',
            name='quantity',
            field=models.BigIntegerField(help_text='How many units (as defined in the associated Ledger instance) this Transaction represents.'),
        ),
        migrations.AlterField(
            model_name='historicaltransaction',
            name='state',
            field=models.CharField(choices=[('created', 'Created'), ('pending', 'Pending'), ('committed', 'Committed'), ('failed', 'Failed')], db_index=True, default='created', help_text="The current state of the Transaction. One of: ['Created', 'Pending', 'Committed', 'Failed']", max_length=255),
        ),
        migrations.AlterField(
            model_name='ledger',
            name='idempotency_key',
            field=models.CharField(blank=True, help_text='An idempotency key is a unique value generated by the client of the Ledger API which the Ledger and Transaction APIs use to recognize subsequent retries of the same redemption (i.e. a Transaction that leads to the fulfillment of an enrollment or entitlement for a particular user id in a particular content key.We suggest incorporating V4 UUIDs along with sufficiently unique or random values representing the desired redemption to avoid unintended collisions.Utility methods are provided in the ``utils.py`` module to help clients generate appropriate idempotency keys.', max_length=255, unique=True),
        ),
        migrations.AlterField(
            model_name='ledger',
            name='metadata',
            field=jsonfield.fields.JSONField(blank=True, help_text='Any additionaly metadata that a client may want to associate with this Ledger instance.', null=True),
        ),
        migrations.AlterField(
            model_name='ledger',
            name='unit',
            field=models.CharField(choices=[('usd_cents', 'U.S. Dollar (Cents)'), ('seats', 'Seats in a course')], db_index=True, default='usd_cents', help_text='The unit in which Transaction quantities related to this Ledger instance are denominated.', max_length=255),
        ),
        migrations.AlterField(
            model_name='reversal',
            name='idempotency_key',
            field=models.CharField(db_index=True, help_text="An idempotency key is a unique value generated by the client of the Ledger API which the Ledger and Transaction APIs use to recognize subsequent retries of the same redemption (i.e. a Transaction that leads to the fulfillment of an enrollment or entitlement for a particular user id in a particular content key).We suggest incorporating V4 UUIDs along with sufficiently unique or random values representing the desired redemption to avoid unintended collisions.In particular, a Transaction's idempotency_key should incorporate it's corresponding ledger's idempotency_key.Utility methods are provided in the ``utils.py`` module to help clients generate appropriate idempotency keys.", max_length=255),
        ),
        migrations.AlterField(
            model_name='reversal',
            name='metadata',
            field=jsonfield.fields.JSONField(blank=True, help_text='Any additionaly metadata that a client may want to associate with this Transaction.', null=True),
        ),
        migrations.AlterField(
            model_name='reversal',
            name='quantity',
            field=models.BigIntegerField(help_text='How many units (as defined in the associated Ledger instance) this Transaction represents.'),
        ),
        migrations.AlterField(
            model_name='reversal',
            name='state',
            field=models.CharField(choices=[('created', 'Created'), ('pending', 'Pending'), ('committed', 'Committed'), ('failed', 'Failed')], db_index=True, default='created', help_text="The current state of the Transaction. One of: ['Created', 'Pending', 'Committed', 'Failed']", max_length=255),
        ),
        migrations.AlterField(
            model_name='reversal',
            name='transaction',
            field=models.OneToOneField(help_text='The Transaction instance which is reversed by this Reversal instance.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='reversal', to='openedx_ledger.transaction'),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='idempotency_key',
            field=models.CharField(db_index=True, help_text="An idempotency key is a unique value generated by the client of the Ledger API which the Ledger and Transaction APIs use to recognize subsequent retries of the same redemption (i.e. a Transaction that leads to the fulfillment of an enrollment or entitlement for a particular user id in a particular content key).We suggest incorporating V4 UUIDs along with sufficiently unique or random values representing the desired redemption to avoid unintended collisions.In particular, a Transaction's idempotency_key should incorporate it's corresponding ledger's idempotency_key.Utility methods are provided in the ``utils.py`` module to help clients generate appropriate idempotency keys.", max_length=255),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='ledger',
            field=models.ForeignKey(help_text='The Ledger instance with which this Transaction is associated.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='transactions', to='openedx_ledger.ledger'),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='lms_user_id',
            field=models.IntegerField(blank=True, db_index=True, help_text='The id of the Open edX LMS user record with which this Transaction is associated.', null=True),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='metadata',
            field=jsonfield.fields.JSONField(blank=True, help_text='Any additionaly metadata that a client may want to associate with this Transaction.', null=True),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='quantity',
            field=models.BigIntegerField(help_text='How many units (as defined in the associated Ledger instance) this Transaction represents.'),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='state',
            field=models.CharField(choices=[('created', 'Created'), ('pending', 'Pending'), ('committed', 'Committed'), ('failed', 'Failed')], db_index=True, default='created', help_text="The current state of the Transaction. One of: ['Created', 'Pending', 'Committed', 'Failed']", max_length=255),
        ),
        migrations.AlterIndexTogether(
            name='transaction',
            index_together={('ledger', 'subsidy_access_policy_uuid'), ('ledger', 'content_key'), ('ledger', 'lms_user_id', 'content_key')},
        ),
    ]



"""
Tests for the openedx_ledger Python API.
"""
import uuid

import pytest

from openedx_ledger import api
from openedx_ledger.models import AdjustmentReasonChoices, Deposit, Transaction, TransactionStateChoices, UnitChoices
from openedx_ledger.test_utils.factories import SalesContractReferenceProviderFactory


@pytest.mark.django_db
def test_create_ledger_happy_path():
    sales_reference_kwargs = {
        "sales_contract_reference_id": str(uuid.uuid4()),
        "sales_contract_reference_provider": SalesContractReferenceProviderFactory(),
    }

    ledger = api.create_ledger(unit=UnitChoices.USD_CENTS, idempotency_key='my-happy-ledger', **sales_reference_kwargs)
    assert ledger.balance() == 0

    api.create_transaction(ledger, quantity=5000, idempotency_key='tx-1', state=TransactionStateChoices.CREATED)
    assert ledger.balance() == 5000

    tx_2 = api.create_transaction(ledger, quantity=5000, idempotency_key='tx-2', state=TransactionStateChoices.CREATED)
    assert ledger.balance() == 10000

    tx_2.state = TransactionStateChoices.COMMITTED
    tx_2.save()

    api.reverse_full_transaction(tx_2, idempotency_key='reversal-1')
    assert ledger.balance() == 5000

    other_ledger = api.create_ledger(unit=UnitChoices.USD_CENTS, idempotency_key='my-happy-ledger')
    assert ledger == other_ledger


@pytest.mark.django_db
def test_create_ledger_with_initial_deposit():
    """
    Simple test case to make sure ledger creation API is capable of creating initial Deposit and Transaction objects.
    """
    ledger = api.create_ledger(
        unit=UnitChoices.USD_CENTS,
        idempotency_key='my-happy-ledger',
        initial_deposit=100,
        sales_contract_reference_id=str(uuid.uuid4()),
        sales_contract_reference_provider=SalesContractReferenceProviderFactory(),
    )
    assert ledger.balance() == 100
    assert ledger.total_deposits() == 100

    # Check that a transaction was created.
    assert len(Transaction.objects.filter(ledger=ledger)) == 1
    assert Transaction.objects.filter(ledger=ledger)[0].quantity == 100

    # Check that a deposit was created.
    assert len(Deposit.objects.filter(ledger=ledger)) == 1
    assert Deposit.objects.filter(ledger=ledger)[0].desired_deposit_quantity == 100


@pytest.mark.django_db
def test_create_ledger_missing_sales_contract_reference():
    """
    Ledger creation API fails if initial_deposit is non-zero but no sales contract reference is provided.
    """
    with pytest.raises(
        api.LedgerCreationError,
        match="both sales_contract_reference_id/provider are required when creating an initial deposit"
    ):
        api.create_ledger(
            unit=UnitChoices.USD_CENTS,
            idempotency_key='my-happy-ledger',
            initial_deposit=100,
        )


@pytest.mark.django_db
def test_no_negative_balance():
    ledger = api.create_ledger(unit=UnitChoices.USD_CENTS, idempotency_key='my-other-ledger')
    assert ledger.balance() == 0

    with pytest.raises(
        api.LedgerBalanceExceeded,
        match="A Transaction was not created because it would exceed the ledger balance."
    ):
        api.create_transaction(ledger, quantity=-1, idempotency_key='tx-1', state=TransactionStateChoices.CREATED)

    api.create_transaction(ledger, quantity=999, idempotency_key='tx-2', state=TransactionStateChoices.CREATED)
    with pytest.raises(
        api.LedgerBalanceExceeded,
        match="A Transaction was not created because it would exceed the ledger balance."
    ):
        api.create_transaction(ledger, quantity=-1000, idempotency_key='tx-3', state=TransactionStateChoices.CREATED)


@pytest.mark.django_db
def test_multiple_reversals():
    ledger = api.create_ledger(unit=UnitChoices.USD_CENTS, idempotency_key='my-other-ledger')
    assert ledger.balance() == 0

    tx_1 = api.create_transaction(ledger, quantity=5000, idempotency_key='tx-1', state=TransactionStateChoices.CREATED)
    assert ledger.balance() == 5000

    tx_1.state = TransactionStateChoices.COMMITTED
    tx_1.save()

    reversal = api.reverse_full_transaction(tx_1, idempotency_key='reversal-1')
    assert ledger.balance() == 0

    with pytest.raises(Exception):
        api.reverse_full_transaction(tx_1, idempotency_key='reversal-2')

    third_reversal = api.reverse_full_transaction(tx_1, idempotency_key='reversal-1')
    assert ledger.balance() == 0
    assert reversal == third_reversal


@pytest.mark.django_db
def test_adjustment_creation_happy_path():
    ledger = api.create_ledger(unit=UnitChoices.USD_CENTS, idempotency_key='my-happy-ledger')
    assert ledger.balance() == 0

    tx_of_interest = api.create_transaction(
        ledger,
        quantity=5000,
        idempotency_key='tx_of_interest',
        state=TransactionStateChoices.COMMITTED,
    )

    assert ledger.balance() == 5000

    test_uuid = uuid.uuid4()

    adjustment = api.create_adjustment(
        ledger,
        adjustment_uuid=test_uuid,
        quantity=-100,
        idempotency_key='unique-string-for-transaction',
        reason=AdjustmentReasonChoices.POOR_CONTENT_FIT,
        notes='Long form notes about this record',
        transaction_of_interest=tx_of_interest,
        some_key='some_value',  # tests that metadata is recorded on the adjustment's transaction
    )

    assert adjustment.uuid == test_uuid
    assert adjustment.transaction.state == TransactionStateChoices.COMMITTED
    assert adjustment.transaction.uuid != tx_of_interest.uuid
    assert adjustment.transaction.idempotency_key == 'unique-string-for-transaction'
    assert adjustment.adjustment_quantity == -100
    assert adjustment.reason == AdjustmentReasonChoices.POOR_CONTENT_FIT
    assert adjustment.notes == 'Long form notes about this record'
    assert adjustment.transaction_of_interest == tx_of_interest
    assert adjustment.transaction.metadata == {
        'some_key': 'some_value',
    }
    assert ledger.balance() == 4900

    # This function is unfortunately not idempotent. But at least it rolls back any unwanted changes and maintains
    # database integrity.
    with pytest.raises(
        api.AdjustmentCreationError,
        match="UNIQUE constraint failed: openedx_ledger_adjustment",
    ):
        api.create_adjustment(
            ledger,
            adjustment_uuid=test_uuid,
            quantity=-100,
            idempotency_key='unique-string-for-transaction',
            reason=AdjustmentReasonChoices.POOR_CONTENT_FIT,
            notes='Long form notes about this record',
            transaction_of_interest=tx_of_interest,
            some_key='some_value',  # tests that metadata is recorded on the adjustment's transaction
        )

    # No change after 2nd attempt.
    assert ledger.balance() == 4900


@pytest.mark.django_db
def test_adjustment_creation_balance_exceeded():
    ledger = api.create_ledger(unit=UnitChoices.USD_CENTS, idempotency_key='my-happy-ledger')

    # Add a little bit of adjustment balance first
    first_adjustment = api.create_adjustment(
        ledger,
        quantity=50,
        reason=AdjustmentReasonChoices.POOR_CONTENT_FIT,
        notes='Long form notes about this record',
    )
    assert ledger.balance() == 50
    assert f'{ledger.uuid}' in str(first_adjustment.transaction.idempotency_key)

    with pytest.raises(
        api.AdjustmentCreationError,
        match="A Transaction was not created because it would exceed the ledger balance."
    ):
        api.create_adjustment(
            ledger,
            quantity=-100,
            reason=AdjustmentReasonChoices.POOR_CONTENT_FIT,
            notes='Long form notes about this record',
        )
    assert ledger.balance() == 50
    assert ledger.transactions.all().count() == 1


@pytest.mark.django_db
@pytest.mark.parametrize(
    "idempotency_key,expected_idempotency_key_substring",
    [
        ('foo-bar', 'foo-bar'),
        (None, 'sales-contract'),
    ],
)
def test_deposit_creation_happy_path(idempotency_key, expected_idempotency_key_substring):
    ledger = api.create_ledger(unit=UnitChoices.USD_CENTS, idempotency_key='my-happy-ledger')
    assert ledger.balance() == 0

    test_deposit_uuid = uuid.uuid4()
    test_sales_contract_reference_id = str(uuid.uuid4())
    test_sales_contract_reference_provider = SalesContractReferenceProviderFactory()

    deposit = api.create_deposit(
        ledger,
        deposit_uuid=test_deposit_uuid,
        quantity=100,
        sales_contract_reference_id=test_sales_contract_reference_id,
        sales_contract_reference_provider=test_sales_contract_reference_provider,
        idempotency_key=idempotency_key,
        some_key='some_value',  # tests that metadata is recorded on the adjustment's transaction
    )

    assert deposit.uuid == test_deposit_uuid
    assert deposit.transaction.quantity == 100
    assert deposit.transaction.state == TransactionStateChoices.COMMITTED
    assert expected_idempotency_key_substring in deposit.transaction.idempotency_key
    assert deposit.desired_deposit_quantity == 100
    assert deposit.sales_contract_reference_id == test_sales_contract_reference_id
    assert deposit.sales_contract_reference_provider == test_sales_contract_reference_provider
    assert deposit.transaction.metadata == {
        'some_key': 'some_value',
    }
    assert ledger.balance() == 100

    # This function is unfortunately not idempotent. But at least it rolls back any unwanted changes and maintains
    # database integrity.
    with pytest.raises(
        api.DepositCreationError,
        match="UNIQUE constraint failed: openedx_ledger_deposit.transaction_id",
    ):
        api.create_deposit(
            ledger,
            deposit_uuid=test_deposit_uuid,
            quantity=100,
            sales_contract_reference_id=test_sales_contract_reference_id,
            sales_contract_reference_provider=test_sales_contract_reference_provider,
            idempotency_key=idempotency_key,
            some_key='some_value',  # tests that metadata is recorded on the adjustment's transaction
        )

    # No change after 2nd attempt.
    assert ledger.balance() == 100


@pytest.mark.django_db
def test_deposit_creation_error():
    ledger = api.create_ledger(unit=UnitChoices.USD_CENTS, idempotency_key='my-happy-ledger')
    assert ledger.balance() == 0

    test_deposit_uuid = uuid.uuid4()
    test_sales_contract_reference_id = str(uuid.uuid4())
    test_sales_contract_reference_provider = SalesContractReferenceProviderFactory()

    with pytest.raises(
        api.DepositCreationError,
        match="Deposits must be positive"
    ):
        api.create_deposit(
            ledger,
            deposit_uuid=test_deposit_uuid,
            quantity=-100,  # negative deposits not allowed, so we expect this to throw an exception.
            sales_contract_reference_id=test_sales_contract_reference_id,
            sales_contract_reference_provider=test_sales_contract_reference_provider,
            idempotency_key='unique-string-for-transaction',
            some_key='some_value',  # tests that metadata is recorded on the adjustment's transaction
        )

    # Make sure the ledger balance did not change.
    assert ledger.balance() == 0



# Generated by Django 3.2.18 on 2023-10-24 19:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openedx_ledger', '0008_adjustment_model'),
    ]

    operations = [
        migrations.AddField(
            model_name='historicaltransaction',
            name='content_title',
            field=models.CharField(blank=True, db_index=True, help_text='The title of the content associated with this Transaction.The title is captured at the time the Transaction is created and may not be up to date.', max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='historicaltransaction',
            name='lms_user_email',
            field=models.CharField(blank=True, db_index=True, help_text='The email of the Open edX LMS user record with which this Transaction is associated.The email is captured at the time the Transaction is created and may not be up to date.', max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='transaction',
            name='content_title',
            field=models.CharField(blank=True, db_index=True, help_text='The title of the content associated with this Transaction.The title is captured at the time the Transaction is created and may not be up to date.', max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='transaction',
            name='lms_user_email',
            field=models.CharField(blank=True, db_index=True, help_text='The email of the Open edX LMS user record with which this Transaction is associated.The email is captured at the time the Transaction is created and may not be up to date.', max_length=255, null=True),
        ),
    ]



# Generated by Django 3.2.18 on 2023-04-04 17:44

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):

    dependencies = [
        ('openedx_ledger', '0005_help_text_and_more_indices'),
    ]

    operations = [
        migrations.CreateModel(
            name='ExternalFulfillmentProvider',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(
                    default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(
                    default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(db_index=True,
                 help_text='The name of the external reference type.', max_length=255, unique=True)),
                ('slug', models.SlugField(
                    help_text='The slug of the external reference type. This is typically the slugified name of the system that the reference is associated with.', max_length=32, unique=True)),
            ],
            options={
                'verbose_name': 'External Fulfillment provider',
                'verbose_name_plural': 'External fulfillment providers',
            },
        ),
        migrations.RemoveField(
            model_name='historicaltransaction',
            name='reference_id',
        ),
        migrations.RemoveField(
            model_name='historicaltransaction',
            name='reference_type',
        ),
        migrations.RemoveField(
            model_name='transaction',
            name='reference_id',
        ),
        migrations.RemoveField(
            model_name='transaction',
            name='reference_type',
        ),
        migrations.AddField(
            model_name='historicaltransaction',
            name='fulfillment_identifier',
            field=models.CharField(blank=True, db_index=True, help_text='The UUID identifier of the subsidized enrollment record for a learner generated durning the enrollmentcall made to enterprise/edx platform e.g. a LearnerCreditEnterpriseCourseEnrollment UUID.', max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='transaction',
            name='fulfillment_identifier',
            field=models.CharField(blank=True, db_index=True, help_text='The UUID identifier of the subsidized enrollment record for a learner generated durning the enrollmentcall made to enterprise/edx platform e.g. a LearnerCreditEnterpriseCourseEnrollment UUID.', max_length=255, null=True),
        ),
        migrations.CreateModel(
            name='ExternalTransactionReference',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(
                    default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(
                    default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('external_reference_id', models.CharField(db_index=True,
                 help_text='The identifier for the external reference operation. This is typically the name of the system that the reference is associated with.', max_length=255, primary_key=True, serialize=False)),
                ('external_fulfillment_provider', models.ForeignKey(help_text='The provider (source) of the external reference.',
                 on_delete=django.db.models.deletion.CASCADE, related_name='reference', to='openedx_ledger.externalfulfillmentprovider')),
                ('transaction', models.ForeignKey(help_text='The Transaction to which this external reference is associated.', null=True,
                 on_delete=django.db.models.deletion.CASCADE, related_name='external_reference', to='openedx_ledger.transaction')),
            ],
            options={
                'verbose_name': 'External Transaction Reference',
                'unique_together': {('external_reference_id', 'external_fulfillment_provider')},
            },
        ),
    ]



# Generated by Django 3.2.16 on 2023-02-14 12:49

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import jsonfield.fields
import model_utils.fields
import simple_history.models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('openedx_ledger', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='HistoricalTransaction',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', models.UUIDField(db_index=True, default=uuid.uuid4, editable=False)),
                ('idempotency_key', models.CharField(db_index=True, max_length=255)),
                ('quantity', models.BigIntegerField()),
                ('metadata', jsonfield.fields.JSONField(blank=True, null=True)),
                ('lms_user_id', models.IntegerField(blank=True, db_index=True, null=True)),
                ('content_uuid', models.UUIDField(blank=True, db_index=True, null=True)),
                ('reference_id', models.CharField(blank=True, db_index=True, help_text='The identifier of the item acquired via the transaction.e.g. a course enrollment ID, an entitlement ID, a subscription license ID.', max_length=255, null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('ledger', models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='openedx_ledger.ledger')),
            ],
            options={
                'verbose_name': 'historical transaction',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalReversal',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', models.UUIDField(db_index=True, default=uuid.uuid4, editable=False)),
                ('idempotency_key', models.CharField(db_index=True, max_length=255)),
                ('quantity', models.BigIntegerField()),
                ('metadata', jsonfield.fields.JSONField(blank=True, null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('transaction', models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='openedx_ledger.transaction')),
            ],
            options={
                'verbose_name': 'historical reversal',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalLedger',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', models.UUIDField(db_index=True, default=uuid.uuid4, editable=False)),
                ('idempotency_key', models.CharField(db_index=True, max_length=255)),
                ('unit', models.CharField(choices=[('usd_cents', 'U.S. Dollar (Cents)'), ('seats', 'Seats in a course'), ('jpy', 'Japanese Yen')], db_index=True, default='usd_cents', max_length=255)),
                ('metadata', jsonfield.fields.JSONField(blank=True, null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical ledger',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
    ]



# Generated by Django 3.2.18 on 2023-03-13 18:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openedx_ledger', '0003_field_updates_20230216_1605'),
    ]

    operations = [
        migrations.AlterField(
            model_name='historicalledger',
            name='idempotency_key',
            field=models.CharField(blank=True, db_index=True, max_length=255),
        ),
        migrations.AlterField(
            model_name='ledger',
            name='idempotency_key',
            field=models.CharField(blank=True, max_length=255, unique=True),
        ),
    ]



# Generated by Django 3.2.16 on 2022-12-14 10:03

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import jsonfield.fields
import model_utils.fields
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Ledger',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, unique=True)),
                ('idempotency_key', models.CharField(max_length=255, unique=True)),
                ('unit', models.CharField(choices=[('usd_cents', 'U.S. Dollar (Cents)'), ('seats', 'Seats in a course'), ('jpy', 'Japanese Yen')], db_index=True, default='usd_cents', max_length=255)),
                ('metadata', jsonfield.fields.JSONField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Transaction',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, unique=True)),
                ('idempotency_key', models.CharField(db_index=True, max_length=255)),
                ('quantity', models.BigIntegerField()),
                ('metadata', jsonfield.fields.JSONField(blank=True, null=True)),
                ('lms_user_id', models.IntegerField(blank=True, db_index=True, null=True)),
                ('content_uuid', models.UUIDField(blank=True, db_index=True, null=True)),
                ('reference_id', models.CharField(blank=True, db_index=True, help_text='The identifier of the item acquired via the transaction.e.g. a course enrollment ID, an entitlement ID, a subscription license ID.', max_length=255, null=True)),
                ('ledger', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='transactions', to='openedx_ledger.ledger')),
            ],
            options={
                'unique_together': {('ledger', 'idempotency_key')},
            },
        ),
        migrations.CreateModel(
            name='Reversal',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, unique=True)),
                ('idempotency_key', models.CharField(db_index=True, max_length=255)),
                ('quantity', models.BigIntegerField()),
                ('metadata', jsonfield.fields.JSONField(blank=True, null=True)),
                ('transaction', models.OneToOneField(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='reversal', to='openedx_ledger.transaction')),
            ],
            options={
                'unique_together': {('transaction', 'idempotency_key')},
            },
        ),
    ]



"""
Test utilities.

Since pytest discourages putting __init__.py into testdirectory
(i.e. making tests a package) one cannot import from anywhere
under tests folder. However, some utility classes/methods might be useful
in multiple test modules (i.e. factoryboy factories, base test classes).

So this package is the place to put them.
"""



"""
edx_ledger models.
"""
from contextlib import contextmanager
from uuid import uuid4

from django.core.cache import cache as django_cache
from django.db import models
from django.db.models import Q
from django.db.models.functions import Coalesce
from django.db.transaction import atomic
from edx_django_utils.cache.utils import get_cache_key
from jsonfield.fields import JSONField
from model_utils.models import TimeStampedModel
from simple_history.models import HistoricalRecords

from openedx_ledger.utils import create_idempotency_key_for_ledger

LEDGER_LOCK_RESOURCE_NAME = 'ledger'


class UnitChoices:
    USD_CENTS = 'usd_cents'
    SEATS = 'seats'
    CHOICES = (
        (USD_CENTS, 'U.S. Dollar (Cents)'),
        (SEATS, 'Seats in a course'),
    )


class TransactionStateChoices:
    """
    Lifecycle states for a ledger transaction.

    CREATED
        Indicates that the transaction has only just been created, and should be the default state.

    PENDING
        Indicates that an attempt is being made to redeem the content in the target LMS.

    COMMITTED
        Indicates that the content has been redeemed, and a reference to the redemption result (often an enrollment ID)
        is stored in the reference_id field of the transaction.
    """

    CREATED = 'created'
    PENDING = 'pending'
    COMMITTED = 'committed'
    FAILED = 'failed'
    CHOICES = (
        (CREATED, 'Created'),
        (PENDING, 'Pending'),
        (COMMITTED, 'Committed'),
        (FAILED, 'Failed'),
    )


class AdjustmentReasonChoices:
    """
    Allowed choices for the ``Adjustment.reason`` field.
    """
    UNAUTHROZIED_ENROLLMENT = 'unauthorized_enrollment'
    POOR_CONTENT_FIT = 'poor_content_fit'
    TECHNICAL_CHALLENGES = 'technical_challenges'
    MISSED_REFUND_OR_DATE = 'missed_refund_or_date'
    GOOD_FAITH = 'good_faith'
    CHOICES = (
        (UNAUTHROZIED_ENROLLMENT, 'Unauthorized enrollment'),
        (POOR_CONTENT_FIT, 'Poor content fit'),
        (TECHNICAL_CHALLENGES, 'Technical challenges'),
        (MISSED_REFUND_OR_DATE, 'Missed refund or date'),
        (GOOD_FAITH, 'Good faith/Relationship building'),
    )


class LedgerLockAttemptFailed(Exception):
    """
    Raise when attempt to lock Ledger failed due to an already existing lock.
    """


class TimeStampedModelWithUuid(TimeStampedModel):
    """
    Base timestamped model adding a UUID field.
    """

    class Meta:
        """
        Metaclass for TimeStampedModelWithUuid.
        """

        abstract = True

    uuid = models.UUIDField(
        primary_key=True,
        default=uuid4,
        editable=False,
        unique=True,
    )


class Ledger(TimeStampedModelWithUuid):
    """
    A ledger to which you can add or remove value, associated with a single subsidy plan.

    All value quantities associated with this ledger uniformly share the same unit, established in the `unit` field of
    the ledger.  This enables simple balancing using `sum()` functions, helping to prevent bugs caused by unit
    conversions, and improving performance on the critical balance() method.

    .. no_pii:
    """
    idempotency_key = models.CharField(
        max_length=255,  # https://docs.djangoproject.com/en/3.2/ref/databases/#mysql-character-fields
        blank=True,
        null=False,
        unique=True,
        help_text=(
            'An idempotency key is a unique value generated by the client of the Ledger API '
            'which the Ledger and Transaction APIs use to recognize subsequent retries '
            'of the same redemption (i.e. a Transaction that leads to the fulfillment of '
            'an enrollment or entitlement for a particular user id in a particular content key.'
            'We suggest incorporating V4 UUIDs along with sufficiently unique or random '
            'values representing the desired redemption to avoid unintended collisions.'
            'Utility methods are provided in the ``utils.py`` module to help clients generate appropriate '
            'idempotency keys.'
        ),
    )
    unit = models.CharField(
        max_length=255,
        blank=False,
        null=False,
        choices=UnitChoices.CHOICES,
        default=UnitChoices.USD_CENTS,
        db_index=True,
        help_text=(
            'The unit in which Transaction quantities related to this Ledger instance are denominated.'
        ),
    )
    metadata = JSONField(
        blank=True,
        null=True,
        help_text=(
            'Any additionaly metadata that a client may want to associate with this Ledger instance.'
        )
    )
    history = HistoricalRecords()

    def subset_balance(self, transactions_queryset):
        """
        Calculate the current balance of the ledger, optionally on a subset of transactions.

        WARNING: The queryset must be a strict subset of self.transactions.  If not, then the return value will
        represent only the set intersection of the given transactions_queryset and self.transactions.

        Args:
            transaction_subset (queryset of openedx_ledger.models.Transaction):
                Transactions to evaluate for the balance calculation.

        Returns:
            int: The total balance of all or a subset of transactions in this ledger.  Possibly a negative value if the
            only transactions evaluated are ones that represent un-reversed enrollment fulfillments.
        """
        with atomic():
            transactions_queryset = transactions_queryset.filter(ledger=self).select_related('reversal')
            agg = transactions_queryset.annotate(
                row_total=Coalesce(models.Sum('quantity'), 0) + Coalesce(models.Sum('reversal__quantity'), 0)
            )
            agg = agg.aggregate(total_quantity=Coalesce(models.Sum('row_total'), 0))
            return agg['total_quantity']

    def balance(self, committed_only=False):
        """
        Calculate the current balance of the ledger.

        TODO: propagate committed_only down into the reversals too.

        Args:
            committed_only (boolean): If true, computes balance only for `committed` transations.  Defaults to False.

        Returns:
            int: The total balance of non-failed transactions in this ledger.
        """
        if committed_only:
            queryset = Transaction.objects.filter(ledger=self, state=TransactionStateChoices.COMMITTED)
        else:
            queryset = Transaction.objects.filter(ledger=self).exclude(state=TransactionStateChoices.FAILED)
        return self.subset_balance(queryset)

    def total_deposits(self):
        """
        Calculate and return the total value added to this ledger.

        Note that this not only factors actual deposits, but also all adjustments which essentially represent
        ad-hoc manipulations of the total value added.

        IMPORTANT NOTE: Until ENT-9132 is completed (backfill initial deposits), this method will be inaccurate in prod.
        Do not call this method until completing ENT-9132!
        """
        # Custom filter that searches for anything that looks and smells like a deposit, including actual deposits.
        deposit_esque_filter = Q(deposit__isnull=False) | Q(adjustment__isnull=False)
        # Limit results to just deposits committed to this ledger.
        queryset = Transaction.objects.filter(
            deposit_esque_filter,
            ledger=self,
            state=TransactionStateChoices.COMMITTED,
        )
        # Finally, subsidy_blance further factors in reversals.
        return self.subset_balance(queryset)

    @property
    def lock_resource_key(self) -> str:
        return get_cache_key(resource=LEDGER_LOCK_RESOURCE_NAME, uuid=self.uuid)

    def acquire_lock(self) -> str:
        """
        Acquire an exclusive lock on this Ledger instance.

        Memcached devs recommend using add() for locking instead of get()+set(), which rules out TieredCache which only
        exposes get()+set() from django cache.  See: https://github.com/memcached/memcached/issues/163

        Returns:
            str: lock ID if a Ledger lock was successfully acquired, None otherwise.
        """
        lock_id = uuid4()
        if django_cache.add(self.lock_resource_key, lock_id):
            return lock_id
        else:
            return None

    def release_lock(self) -> None:
        """
        Release an exclusive lock on this Ledger instance.
        """
        django_cache.delete(self.lock_resource_key)

    @contextmanager
    def lock(self):
        """
        Context manager for locking this Ledger instance.

        Raises:
            openedx_ledger.models.LedgerLockAttemptFailed:
                Raises this if there's another distributed process locking this Ledger.
        """
        lock_id = self.acquire_lock()
        if not lock_id:
            raise LedgerLockAttemptFailed(f"Failed to acquire lock <{lock_id}> on Ledger {str(self)}.")
        try:
            yield lock_id
        finally:
            self.release_lock()

    def save(self, *args, **kwags):
        """
        Sets the idempotency_key for the ledger if it is currently null.
        """
        if not self.idempotency_key:
            self.idempotency_key = create_idempotency_key_for_ledger()
        super().save(*args, **kwags)

    def __str__(self):
        """
        Return string representation of this ledger, visible in logs, django admin, etc.
        """
        return f'<Ledger uuid={self.uuid},\nidempotency_key={self.idempotency_key}>'


class BaseTransaction(TimeStampedModelWithUuid):
    """
    Base class for all models that resemble transactions.
    """

    class Meta:
        """
        Metaclass for BaseTransaction.
        """

        abstract = True

    idempotency_key = models.CharField(
        max_length=255,
        blank=False,
        null=False,
        db_index=True,
        help_text=(
            "An idempotency key is a unique value generated by the client of the Ledger API "
            "which the Ledger and Transaction APIs use to recognize subsequent retries "
            "of the same redemption (i.e. a Transaction that leads to the fulfillment of "
            "an enrollment or entitlement for a particular user id in a particular content key)."
            "We suggest incorporating V4 UUIDs along with sufficiently unique or random "
            "values representing the desired redemption to avoid unintended collisions."
            "In particular, a Transaction's idempotency_key should incorporate it's corresponding "
            "ledger's idempotency_key."
            "Utility methods are provided in the ``utils.py`` module to help clients generate appropriate "
            "idempotency keys."
        ),
    )
    quantity = models.BigIntegerField(
        null=False,
        blank=False,
        help_text=(
            "How many units (as defined in the associated Ledger instance) this Transaction represents."
        ),
    )
    metadata = JSONField(
        blank=True,
        null=True,
        help_text=(
            "Any additionaly metadata that a client may want to associate with this Transaction."
        ),
    )
    state = models.CharField(
        max_length=255,
        blank=False,
        null=False,
        choices=TransactionStateChoices.CHOICES,
        default=TransactionStateChoices.CREATED,
        db_index=True,
        help_text=(
            "The current state of the Transaction. One of: "
            f"{[choice[1] for choice in TransactionStateChoices.CHOICES]}"
        ),
    )


class Transaction(BaseTransaction):
    """
    Represents value moving in or out of the ledger.

    Transactions (and reversals) are immutable after entering the committed state.  This immutability helps maintain a
    complete and robust history of value changes to the ledger, a trait which we rely on to calculate the ledger
    balance.

    Relatedly, we intentionally avoid persisting aggregates, reinforcing the transactions themselves as the
    only source of truth.

    .. no_pii:
    """

    class Meta:
        """
        Metaclass for Transaction.
        """

        unique_together = [('ledger', 'idempotency_key')]
        index_together = [
            ('ledger', 'lms_user_id', 'content_key'),
            ('ledger', 'content_key'),
            ('ledger', 'subsidy_access_policy_uuid'),
        ]

    ledger = models.ForeignKey(
        Ledger,
        related_name='transactions',
        null=True,
        on_delete=models.SET_NULL,
        help_text=(
            "The Ledger instance with which this Transaction is associated."
        )
    )
    lms_user_id = models.IntegerField(
        null=True,
        blank=True,
        db_index=True,
        help_text=(
            "The id of the Open edX LMS user record with which this Transaction is associated."
        ),
    )
    lms_user_email = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        db_index=True,
        help_text=(
            "The email of the Open edX LMS user record with which this Transaction is associated."
            "The email is captured at the time the Transaction is created and may not be up to date."
        )
    )
    content_key = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        db_index=True,
        help_text=(
            "The globally unique content identifier.  Joinable with ContentMetadata.content_key in enterprise-catalog."
        )
    )
    parent_content_key = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        db_index=True,
        help_text=(
            "Identifier for the parent of the content_key.  Also should be joinable with ContentMetadata.content_key "
            "in enterprise-catalog. In practice, this is the course key corresponding to a course run key."
        )
    )
    content_title = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        db_index=True,
        help_text=(
            "The title of the content associated with this Transaction."
            "The title is captured at the time the Transaction is created and may not be up to date."
        )
    )
    fulfillment_identifier = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        db_index=True,
        help_text=(
            "The UUID identifier of the subsidized enrollment record for a learner generated durning the enrollment"
            "call made to enterprise/edx platform e.g. a LearnerCreditEnterpriseCourseEnrollment UUID."
        ),
    )
    subsidy_access_policy_uuid = models.UUIDField(
        blank=True,
        null=True,
        help_text="A reference to the subsidy access policy which was used to create a transaction for the content."
    )
    history = HistoricalRecords()

    def get_reversal(self):
        """
        Convenience method for fetching this transaction's related
        reversal, or None if no such reversal exists.
        """
        try:
            return self.reversal  # pylint: disable=no-member
        except Reversal.DoesNotExist:
            return None

    def get_adjustment(self):
        """
        Convenience method for fetching the ``Adjustment`` which
        created this transaction, if one exists.  If one does not exist,
        returns ``None``.
        """
        try:
            return self.adjustment  # pylint: disable=no-member
        except Adjustment.DoesNotExist:
            return None


class ExternalFulfillmentProvider(TimeStampedModel):
    """
    Model of external fulfillment providers. This is used to track the external systems that are used to fulfill
    transactions.

    .. no_pii:
    """

    class Meta:
        """
        Metaclass for ExternalFulfillmentProvider.
        """

        verbose_name = 'External Fulfillment provider'
        verbose_name_plural = 'External fulfillment providers'

    name = models.CharField(
        max_length=255,
        help_text="The name of the external reference type.",
    )
    slug = models.SlugField(
        max_length=32,
        unique=True,
        help_text=(
            "The slug of the external reference type. This is typically the slugified name of the system that the "
            "reference is associated with."
        )
    )

    def __str__(self):
        """
        Return string representation of this external fulfillment provider, visible in logs, django admin, etc.
        """
        msg = (
            f'<ExternalFulfillmentProvider\n'
            f'name={self.name}\n'
            f'slug={self.slug}>'
        )
        return msg


class ExternalTransactionReference(TimeStampedModel):
    """
    Model of references to transactions from an external system and their associated Transaction objects.

    .. no_pii:
    """

    class Meta:
        """
        Metaclass for ExternalTransactionReference.
        """

        unique_together = [('external_reference_id', 'external_fulfillment_provider')]
        verbose_name = 'External Transaction Reference'

    transaction = models.ForeignKey(
        Transaction,
        related_name='external_reference',
        null=True,
        on_delete=models.CASCADE,
        help_text=(
            "The Transaction to which this external reference is associated."
        ),
    )
    external_reference_id = models.CharField(
        max_length=255,
        blank=False,
        null=False,
        primary_key=True,
        db_index=True,
        help_text=(
            "The identifier for the external reference operation. This is typically the name of the system that the "
            "reference is associated with."
        ),
    )
    external_fulfillment_provider = models.ForeignKey(
        ExternalFulfillmentProvider,
        related_name='reference',
        null=False,
        blank=False,
        on_delete=models.CASCADE,
        help_text=(
            "The provider (source) of the external reference."
        )
    )

    def __str__(self):
        """
        Return string representation of this external reference, visible in logs, django admin, etc.
        """
        msg = (
            f'<ExternalTransactionReference\n'
            f'external_reference_id={self.external_reference_id}\n'
            f'external_fulfillment_provider={self.external_fulfillment_provider}\n'
        )
        if self.transaction:
            msg += f'associated with {self.transaction}\n'
        msg += '>'
        return msg


class Reversal(BaseTransaction):
    """
    Represents a reversal of some or all of a transaction, but no more.

    .. no_pii:
    """

    class Meta:
        """
        Metaclass for Reversal.
        """

        unique_together = [('transaction', 'idempotency_key')]

    transaction = models.OneToOneField(
        Transaction,
        related_name='reversal',
        null=True,
        on_delete=models.SET_NULL,
        help_text=(
            "The Transaction instance which is reversed by this Reversal instance."
        ),
    )
    history = HistoricalRecords()
    # Reversal quantities should always have the opposite sign of the transaction (i.e. negative)
    # We have to enforce this somehow...


class Adjustment(TimeStampedModelWithUuid):
    """
    Represents some adjustment to the balance of a ledger via
    a transaction (with quantity > 0) and some audit fields.

    .. no_pii:
    """
    ledger = models.ForeignKey(
        Ledger,
        related_name='adjustments',
        null=False,
        on_delete=models.CASCADE,
        help_text=(
            "The Ledger instance with which this Adjustment is associated."
        )
    )
    adjustment_quantity = models.BigIntegerField(
        null=False,
        blank=False,
        help_text=(
            "How many units to adjust for in the related transaction."
        ),
    )
    transaction = models.OneToOneField(
        Transaction,
        related_name='adjustment',
        unique=True,
        null=False,
        on_delete=models.CASCADE,
        help_text=(
            "The Transaction instance which adjusts the balance of the relevant ledger."
        ),
    )
    transaction_of_interest = models.OneToOneField(
        Transaction,
        related_name='adjustment_of_interest',
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        help_text=(
            "Any transaction of interest w.r.t. the reason for being of this adjustment. "
            "For example, the transaction of interest may point to some transaction record "
            "for which the enrolling user is unsatisfied, but for which we cannot issue a reversal "
            "due to business rules."
        ),
    )
    reason = models.CharField(
        max_length=255,
        blank=False,
        null=False,
        choices=AdjustmentReasonChoices.CHOICES,
        default=AdjustmentReasonChoices.TECHNICAL_CHALLENGES,
        db_index=True,
        help_text=(
            'The primary reason for the existence of this adjustment.'
        ),
    )
    notes = models.TextField(
        blank=True,
        null=True,
        help_text=(
            'Any additional context you have for the existence of this adjustment.'
        ),
    )
    history = HistoricalRecords()


class SalesContractReferenceProvider(TimeStampedModel):
    """
    All possible providers for sales contracts.

    Typically your sales contracts are recorded within one centralized system-of-record (e.g. SalesForce), in which case
    this table may only ever have one single record to describe that provider.

    .. no_pii:
    """

    class Meta:
        """
        Metaclass for SalesContractReferenceProvider.
        """

        verbose_name = 'Sales Contract Reference Provider'

    name = models.CharField(
        max_length=255,
        help_text="The pretty name of the contract reference provider.",
    )
    slug = models.SlugField(
        max_length=32,
        unique=True,
        help_text="A unique slug for this sales contract reference provider.",
    )

    def __str__(self):
        """
        Return string representation of this external fulfillment provider, visible in logs, django admin, etc.
        """
        msg = (
            f'<SalesContractReferenceProvider\n'
            f'name={self.name}\n'
            f'slug={self.slug}>'
        )
        return msg


class Deposit(TimeStampedModelWithUuid):
    """
    Represents a deposit into to the balance of a ledger via a transaction (with quantity > 0).

    .. no_pii:
    """
    ledger = models.ForeignKey(
        Ledger,
        related_name='deposits',
        null=False,
        on_delete=models.CASCADE,
        help_text=(
            "The Ledger instance with which this deposit is associated."
        )
    )
    # This field is non-contributing, and should NOT be factored into balance calculations. It should only be used to
    # accept input from an admin form and feed the transaction creation.
    desired_deposit_quantity = models.BigIntegerField(
        null=False,
        blank=False,
        help_text=(
            "How many units to deposit via related transaction."
        ),
    )
    transaction = models.OneToOneField(
        Transaction,
        related_name='deposit',
        unique=True,
        null=False,
        on_delete=models.CASCADE,
        help_text=(
            "The Transaction instance which represents this deposit and influences the balance of the relevant ledger."
        ),
    )
    sales_contract_reference_id = models.CharField(
        max_length=255,
        # Don't require this field because it's often inherited from the parent subsidy which also has blank/null=True.
        blank=True,
        null=True,
        db_index=True,
        help_text=(
            "The reference ID for the specific sales contract which beget this deposit. The source system which ",
            "stores the sales object is identified by the sales_contract_reference_provider field.",
        ),
    )
    sales_contract_reference_provider = models.ForeignKey(
        SalesContractReferenceProvider,
        related_name='deposits',
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        help_text=(
            "The system providing the source of truth for the sales contract which beget this deposit."
        )
    )
    history = HistoricalRecords()



"""
Constants for the openedx_ledger app.
"""
CENTS_PER_US_DOLLAR = 100
LEDGER_DEFAULT_IDEMPOTENCY_KEY_PREFIX = 'ledger-default'
LEDGERED_SUBSIDY_IDEMPOTENCY_KEY_PREFIX = 'ledger-for-subsidy'

INITIAL_DEPOSIT_TRANSACTION_SLUG = 'initial-deposit'



"""
Views for the openedx_ledger app.
"""
import logging

from django.core.exceptions import ObjectDoesNotExist
from django.db.utils import IntegrityError
from django.http import HttpResponseBadRequest, HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse
from django.views.generic import View

from openedx_ledger.api import CannotReverseAdjustmentError, NonCommittedTransactionError, reverse_full_transaction
from openedx_ledger.models import Transaction

logger = logging.getLogger(__name__)


class ReverseTransactionView(View):
    """
    Admin view for the form responsible for reversing transaction objects and emitting an unenrollment signal.
    """
    template = "edx_ledger/admin/reverse_transaction.html"

    def get(self, request, transaction_id):
        """
        Handle GET request - render "Reverse Transaction" form.

        Arguments:
            request (django.http.request.HttpRequest): Request instance
            transaction_uuid (str): Enterprise Customer UUID

        Returns:
            django.http.response.HttpResponse: HttpResponse
        """
        transaction = Transaction.objects.get(uuid=transaction_id)
        try:
            if transaction.reversal:
                logger.info(
                    f"ReverseTransactionView Error: transaction reversal already exists: {transaction_id}"
                )
                return HttpResponseBadRequest('Transaction Reversal already exists')
        except ObjectDoesNotExist:
            pass

        return render(
            request,
            self.template,
            {'transaction': transaction}
        )

    def post(self, request, transaction_id):
        """
        Handle POST request - handle form submissions.
        """
        logger.info(
            "Reversing transaction and sending admin invoked transaction unenroll signal for transaction: "
            f"{transaction_id}"
        )
        transaction = Transaction.objects.get(uuid=transaction_id)
        # TODO: determine how to generate idempotency key
        try:
            reverse_full_transaction(
                transaction,
                idempotency_key=f"admin-invoked-reverse-{str(transaction.uuid)}"
            )
        except IntegrityError:
            logger.exception(
                f"ReverseTransactionView Error: transaction reversal already exists: {transaction_id}"
            )
            return HttpResponseBadRequest('Transaction Reversal already exists')
        except NonCommittedTransactionError as error:
            logger.exception(
                f"ReverseTransactionView Error: transaction is not in a committed state: {transaction_id}"
            )
            return HttpResponseBadRequest(f'Transaction Reversal failed: {error}')
        except CannotReverseAdjustmentError as error:
            logger.exception(
                f"ReverseTransactionView Error: cannot reverse the transaction of an adjustment: {transaction_id}"
            )
            return HttpResponseBadRequest(f'Transaction Reversal failed: {error}')
        url = reverse("admin:openedx_ledger_transaction_change", args=(transaction_id,))
        return HttpResponseRedirect(url)



"""
openedx_ledger related signals.
"""

from django.dispatch import Signal

# Signal that indicates that a transaction object has been reversed.
# providing_args=[
#         'reversal',  # Reversal object
#     ]
TRANSACTION_REVERSED = Signal()



# Generated by Django 3.2.18 on 2023-05-01 08:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openedx_ledger', '0006_auto_20230404_1744'),
    ]

    operations = [
        migrations.AlterField(
            model_name='externalfulfillmentprovider',
            name='name',
            field=models.CharField(help_text='The name of the external reference type.', max_length=255),
        ),
    ]



"""
The openedx_ledger python API.
"""
import logging
import uuid

from django.db.transaction import atomic, get_connection

from openedx_ledger import models, utils
from openedx_ledger.signals.signals import TRANSACTION_REVERSED

logger = logging.getLogger(__name__)


class LedgerBalanceExceeded(Exception):
    """
    Exception for when a transaction could not be created because it would exceed the ledger balance.
    """


class NonCommittedTransactionError(Exception):
    """
    Raised when a transaction that is not in a COMMITTED state is used in a reversal.
    """


class CannotReverseAdjustmentError(Exception):
    """
    Raised when a caller attempts to reverse the transaction that comprises
    an ``Adjustment`` record.
    """


class LedgerCreationError(Exception):
    """
    Raised when, for whatever reason, a Ledger could not be created.
    """


class AdjustmentCreationError(Exception):
    """
    Raised when, for whatever reason, an adjustment could not be created.
    """


class DepositCreationError(Exception):
    """
    Raised when, for whatever reason, deposit could not be created.
    """


def create_transaction(
    ledger,
    quantity,
    idempotency_key,
    lms_user_id=None,
    lms_user_email=None,
    content_key=None,
    parent_content_key=None,
    content_title=None,
    subsidy_access_policy_uuid=None,
    state=models.TransactionStateChoices.CREATED,
    **metadata
):
    """
    Create a pending transaction.

    Args:
        ledger (openedx_ledger.models.Ledger): The ledger to which the transaction should be added.
        quantity (int): Negative value representing the desired quantity of the new Transaction.
        idempotency_key (str): The idempotency_key of the new Transaction.
        lms_user_id (int, Optional):
            The lms_user_id representing the learner who is enrolling. Skip if this does not represent a policy
            enrolling a learner into content.
        lms_user_email (str, Optional):
            The lms_user_email representing the learner who is enrolling. Skip if this does not represent a policy
            enrolling a learner into content or if the email is not readily available.
        content_key (str, Optional):
            The identifier of the content into which the learner is enrolling. Skip if this does not represent a policy
            enrolling a learner into content.
        parent_content_key (str, Optional):
            Identifier for the parent of the content_key. Skip if this does not represent a policy enrolling a learner
            into content.
        content_title (str, Optional):
            The title of the content into which the learner is enrolling. Skip if this does not represent a policy
            enrolling a learner into content or if the title is not readily available.
        subsidy_access_policy_uuid (str, Optional):
            The policy which permitted the creation of the new Transaction. Skip if this does not represent a policy
            enrolling a learner into content.
        state (str, Optional):
            The initial state of the new transaction. Choice of openedx_ledger.models.TransactionStateChoices.
        **metadata (dict, Optional):
            Optional metadata to add to the transaction, potentially useful for debugging, analytics, or other purposes
            defined by the caller.

    Raises:
        openedx_ledger.models.LedgerLockAttemptFailed:
            Raises this if there's another attempt in process to add a transaction to this Ledger.
        openedx_ledger.api.LedgerBalanceExceeded:
            Raises this if the transaction would cause the balance of the ledger to become negative.
    """
    with ledger.lock():
        durable = not get_connection().in_atomic_block
        with atomic(durable=durable):
            balance = ledger.balance()
            if (quantity < 0) and ((balance + quantity) < 0):
                raise LedgerBalanceExceeded("A Transaction was not created because it would exceed the ledger balance.")

            transaction, _ = models.Transaction.objects.get_or_create(
                ledger=ledger,
                idempotency_key=idempotency_key,
                defaults={
                    "quantity": quantity,
                    "content_key": content_key,
                    "parent_content_key": parent_content_key,
                    "content_title": content_title,
                    "lms_user_id": lms_user_id,
                    "lms_user_email": lms_user_email,
                    "subsidy_access_policy_uuid": subsidy_access_policy_uuid,
                    "state": state,
                    "metadata": metadata,
                },
            )
    return transaction


def reverse_full_transaction(transaction, idempotency_key, **metadata):
    """
    Reverse a transaction, in full.

    Idempotency of reversals - reversing the same transaction twice
    produces the same output and has no side effect on the second invocation.
    Support idempotency key here, too.
    openedx_ledger.api.NonCommittedTransactionError:
        Raises this if the transaction is not in a COMMITTED state.
    """
    # Do not allow the reversal of transactions that comprise an Adjustment
    if transaction.get_adjustment():
        raise CannotReverseAdjustmentError(
            f"Transaction {transaction.uuid} comprises an Adjustment, can't reverse."
        )

    with atomic(durable=True):
        # select the transaction and any reversals
        # if there is a reversal: return, no work to do here
        # if not, write a reversal for the transaction
        transaction.refresh_from_db()

        if transaction.state != models.TransactionStateChoices.COMMITTED:
            raise NonCommittedTransactionError(
                f"Cannot reverse transaction {transaction.uuid} "
                "because it is not in a committed state."
            )

        reversal, _ = models.Reversal.objects.get_or_create(
            transaction=transaction,
            idempotency_key=idempotency_key,
            defaults={
                'quantity': transaction.quantity * -1,
                'metadata': metadata,
            },
            state=models.TransactionStateChoices.COMMITTED,
        )
        TRANSACTION_REVERSED.send(sender=models.Reversal, reversal=reversal)
        return reversal


def create_ledger(
    unit=None,
    idempotency_key=None,
    subsidy_uuid=None,
    initial_deposit=None,
    sales_contract_reference_id=None,
    sales_contract_reference_provider=None,
    **metadata
):
    """
    Primary interface for creating a Ledger record.

    params:
      unit: Optional unit for the ledger, defaults to the model default of USD_CENTS.
      idempotency_key: Optional idempotency key, defaults to result of
        the utility function for creating ledger idempotency key
      subsidy_uuid: Optional subsidy uuid used in above utility function.
      initial_deposit: Optional amount of value to initialize the ledger with.  In units specified by units
        (units by default are USD_CENTS).
      sales_contract_reference_id (str):
        The reference ID for the specific sales contract which beget this ledger's initial deposit.
      sales_contract_reference_provider (openedx_ledger.models.SalesContractReferenceProvider):
        The system providing the source of truth for the sales contract which beget this ledger's initial deposit.
      metadata: Optional additional information for the ledger.
    """
    if initial_deposit and not (sales_contract_reference_id and sales_contract_reference_provider):
        raise LedgerCreationError(
            "create_ledger(): both sales_contract_reference_id/provider are required when creating an initial deposit"
        )
    ledger, _ = models.Ledger.objects.get_or_create(
        unit=unit or models.UnitChoices.USD_CENTS,
        idempotency_key=idempotency_key or utils.create_idempotency_key_for_ledger(subsidy_uuid),
        defaults={
            'metadata': metadata,
        },
    )
    if initial_deposit:
        initial_idpk = utils.create_idempotency_key_for_transaction(
            ledger,
            initial_deposit,
            is_initial_deposit=True,
        )
        create_deposit(
            ledger=ledger,
            quantity=initial_deposit,
            idempotency_key=initial_idpk,
            sales_contract_reference_id=sales_contract_reference_id,
            sales_contract_reference_provider=sales_contract_reference_provider,
        )

    return ledger


def create_adjustment(
    ledger,
    quantity,
    adjustment_uuid=None,
    reason=models.AdjustmentReasonChoices.TECHNICAL_CHALLENGES,
    notes=None,
    idempotency_key=None,
    transaction_of_interest=None,
    **metadata,
):
    """
    Creates a new Transaction and related Adjustment record
    to adjust the balance of the given ledger.
    """
    if idempotency_key is None:
        tx_idempotency_key = f'{ledger.uuid}-adjustment-{quantity}-reason-{uuid.uuid4()}'
    else:
        tx_idempotency_key = idempotency_key

    try:
        with atomic():
            transaction = create_transaction(
                ledger,
                quantity,
                idempotency_key=tx_idempotency_key,
                state=models.TransactionStateChoices.COMMITTED,
                **metadata,
            )
            kwargs = {}
            if adjustment_uuid:
                kwargs['uuid'] = adjustment_uuid
            adjustment = models.Adjustment.objects.create(
                ledger=ledger,
                adjustment_quantity=quantity,
                transaction=transaction,
                transaction_of_interest=transaction_of_interest,
                reason=reason,
                notes=notes,
                **kwargs,
            )
    except Exception as exc:
        message = f'Failed to create adjustment in ledger {ledger.uuid} for amount {quantity}'
        logger.exception(message)
        raise AdjustmentCreationError(str(exc)) from exc

    return adjustment


def create_deposit(
    ledger,
    quantity,
    sales_contract_reference_id,
    sales_contract_reference_provider,
    deposit_uuid=None,
    idempotency_key=None,
    **metadata,
):
    """
    Creates a new Transaction and related Deposit record to increase the balance of the given ledger.

    Note: this function is NOT idempotent. If called twice with the same idempotency key, it will raise an
    openedx_ledger.api.DepositCreationError. This is probably not ideal.
    """
    if idempotency_key is None:
        tx_idempotency_key = f'{ledger.uuid}-deposit-{quantity}-sales-contract-{sales_contract_reference_id}'
    else:
        tx_idempotency_key = idempotency_key

    if quantity < 0:
        raise DepositCreationError(f'Deposit quantity of {quantity} is invalid becase Deposits must be positive.')

    try:
        with atomic():
            transaction = create_transaction(
                ledger,
                quantity,
                idempotency_key=tx_idempotency_key,
                state=models.TransactionStateChoices.COMMITTED,
                **metadata,
            )
            kwargs = {}
            if deposit_uuid:
                kwargs['uuid'] = deposit_uuid
            deposit = models.Deposit.objects.create(
                ledger=ledger,
                desired_deposit_quantity=quantity,
                transaction=transaction,
                sales_contract_reference_id=sales_contract_reference_id,
                sales_contract_reference_provider=sales_contract_reference_provider,
                **kwargs,
            )
    except Exception as exc:
        message = f'Failed to create desposit in ledger {ledger.uuid} for amount {quantity}'
        logger.exception(message)
        raise DepositCreationError(str(exc)) from exc

    return deposit



#!/usr/bin/env python
"""
Tests for the `openedx-ledger` models.
"""
import uuid

import ddt
import pytest
from django.test import TestCase

from openedx_ledger import models
from openedx_ledger.constants import INITIAL_DEPOSIT_TRANSACTION_SLUG
from openedx_ledger.test_utils.factories import (
    AdjustmentFactory,
    DepositFactory,
    LedgerFactory,
    ReversalFactory,
    TransactionFactory,
)


@ddt.ddt
class LedgerBalanceTests(TestCase):
    """
    Tests for the balance of the Ledger model.
    """

    def setUp(self):
        super().setUp()
        self.ledger = LedgerFactory()
        self.initial_deposit = DepositFactory(
            ledger=self.ledger,
            desired_deposit_quantity=100,
        )
        self.initial_deposit.transaction.idempotency_key = INITIAL_DEPOSIT_TRANSACTION_SLUG
        self.initial_deposit.transaction.save()
        self.transaction_2 = TransactionFactory(
            ledger=self.ledger, lms_user_id=1, content_key="course-v1:edX+test+course.1", quantity=-10
        )
        self.transaction_3 = TransactionFactory(
            ledger=self.ledger, lms_user_id=1, content_key="course-v1:edX+test+course.2", quantity=-10
        )
        self.transaction_4 = TransactionFactory(
            ledger=self.ledger, lms_user_id=2, content_key="course-v1:edX+test+course.2", quantity=-10
        )
        self.reversal_1 = ReversalFactory(
            transaction=self.transaction_4,
            quantity=10,
        )
        self.transaction_5 = TransactionFactory(
            ledger=self.ledger,
            lms_user_id=3,
            lms_user_email='user@example.com',
            content_key="course-v1:edX+test+course.3",
            parent_content_key="edX+test",
            content_title="Edx: test course 3",
            quantity=-10,
            state=models.TransactionStateChoices.PENDING,
        )
        self.adjustment_1 = AdjustmentFactory(
            ledger=self.ledger,
            adjustment_quantity=15,
        )
        self.adjustment_2 = AdjustmentFactory(
            ledger=self.ledger,
            adjustment_quantity=-5,
        )
        self.extra_deposit = DepositFactory(
            ledger=self.ledger,
            desired_deposit_quantity=50,
        )

    def test_balance(self):
        """
        Test ``Ledger.balance()``.

        ``committed_only`` is implicitly False.
        """
        result = self.ledger.balance()
        assert result == 100 - 10 - 10 - 10 + 10 - 10 + 15 - 5 + 50

    def test_balance_committed_only(self):
        """
        Test ``Ledger.balance(committed_only=True)``.

        One of the setUp transactions is still pending.
        """
        result = self.ledger.balance(committed_only=True)
        assert result == 100 - 10 - 10 - 10 + 10 + 15 - 5 + 50

    @ddt.data(
        # Test calculating the balance of the entire ledger.
        {
            "transaction_filters": {},
            "expected_balance": 100 - 10 - 10 - 10 + 10 - 10 + 15 - 5 + 50,
        },
        # Test calculating the balance of a user subset.
        {
            "transaction_filters": {"lms_user_id": 1},
            "expected_balance": -10 - 10,
        },
        # Test calculating the balance of a user subset, reversal case.
        {
            "transaction_filters": {"lms_user_id": 2},
            "expected_balance": -10 + 10,
        },
        # Test calculating the balance of a content subset, with a reversal.
        {
            "transaction_filters": {"content_key": "course-v1:edX+test+course.2"},
            "expected_balance": -10 - 10 + 10,
        },
    )
    @ddt.unpack
    def test_subset_balance(self, transaction_filters, expected_balance):
        """
        Test Ledger.subset_balance().
        """
        queryset = models.Transaction.objects.filter(ledger=self.ledger, **transaction_filters)
        result = self.ledger.subset_balance(queryset)

        assert result == expected_balance

    def test_subset_balance_doesnt_fail_superset(self):
        """
        Test Ledger.subset_balance() doesn't fail when given a set of transactions that are a superset of
        Ledger.transactions.
        """
        assert self.ledger.subset_balance(models.Transaction.objects.all()) == self.ledger.balance()

    def test_balance_excludes_failed_transactions(self):
        """
        Failed transations should not be included in the balance calculation.
        """
        _ = TransactionFactory(
            ledger=self.ledger,
            lms_user_id=2,
            content_key="course-v1:edX+test+course.2",
            quantity=-55,
            state=models.TransactionStateChoices.FAILED,
        )
        expected_balance = 100 - 10 - 10 - 10 + 10 - 10 + 15 - 5 + 50

        self.assertEqual(self.ledger.balance(), expected_balance)

    def test_total_deposits(self):
        """
        Test Ledger.total_deposits() counts all deposit-esque transactions.
        """
        #                       extra deposit─────────────────┐
        #                         adjustments────────┬────┐   │
        #                     initial deposit───┐    │    │   │
        assert self.ledger.total_deposits() == 100 + 15 - 5 + 50
        # Note that all transactions representing learner spend are excluded.

    def test_idempotency_key_is_generated(self):
        """
        Tests that Ledger.save() will create an idempotency_key
        if none exists.
        """
        my_ledger = LedgerFactory(idempotency_key=None)
        my_ledger.save()

        self.assertIsNotNone(my_ledger.idempotency_key)


class LedgerLockTests(TestCase):
    """
    Test the locking mechanisms for ledgers.
    """
    def setUp(self):
        super().setUp()

        self.ledger = LedgerFactory()
        self.other_ledger = LedgerFactory()

    def test_acquire_lock_release_lock(self):
        """
        Create one hypothetical sequence consisting of three actors and two ledgers.  Each Ledger should only allow one
        lock to be grabbed at a time.
        """
        # Simple case, acquire lock on first ledger.
        lock_1 = self.ledger.acquire_lock()
        assert lock_1  # Non-null means the lock was successfully acquired.
        # A second actor attempts to acquire lock on first ledger, but it's already locked.
        lock_2 = self.ledger.acquire_lock()
        assert lock_2 is None
        # A third actor attempts to acquire lock on second ledger, should work even though first ledger is locked.
        lock_3 = self.other_ledger.acquire_lock()
        assert lock_3
        assert lock_3 != lock_1
        # After releasing the first lock, the second actor should have success.
        self.ledger.release_lock()
        lock_2 = self.ledger.acquire_lock()
        assert lock_2
        # Finally, the third actor releases the lock on the second ledger.
        self.other_ledger.release_lock()

    def test_lock_contextmanager_happy(self):
        """
        Ensure the lock contextmanager does not raise an exception if the ledger is not locked.
        """
        with self.ledger.lock():
            pass

    def test_lock_contextmanager_already_locked(self):
        """
        Ensure the lock contextmanager raises LedgerLockAttemptFailed if the ledger is locked.
        """
        self.ledger.acquire_lock()
        with pytest.raises(models.LedgerLockAttemptFailed, match=r"Failed to acquire lock.*"):
            with self.ledger.lock():
                pass


class TestExternalFulfillmentAndReference(TestCase):
    """
    Tests for the ExternalFulillmentProvider and ExternalTransactionReference models.
    """
    @classmethod
    def setUpTestData(cls):
        """
        Set up some test objects.
        """
        super().setUpTestData()

        cls.provider = models.ExternalFulfillmentProvider.objects.create(
            name='My provider',
            slug='my-provider',
        )
        cls.ledger = LedgerFactory()
        cls.initial_transaction = TransactionFactory(
            ledger=cls.ledger,
            quantity=100,
        )

    def test_simple_provider_stuff(self):
        """
        Test we can str() these.
        """
        self.assertIn('my-provider', str(self.provider))
        self.assertIn('My provider', str(self.provider))

    def test_external_reference(self):
        """
        Test that we can create an external transaction reference.
        """
        transaction = TransactionFactory(
            ledger=self.ledger,
        )
        external_id = str(uuid.uuid4())
        external_reference = models.ExternalTransactionReference.objects.create(
            transaction=transaction,
            external_fulfillment_provider=self.provider,
            external_reference_id=external_id,
        )

        self.assertIn(external_id, str(external_reference))
